import unittest
from decimal import (
    ROUND_CEILING,
    ROUND_DOWN,
    ROUND_FLOOR,
    ROUND_HALF_DOWN,
    ROUND_HALF_UP,
    ROUND_UP,
    Decimal,
)
from time import time
from typing import Iterable, Optional
from unittest.mock import Mock

from amarket.base import MarketBase
from amarket.constants import TIMEFRAMES
from test_utils import check_primitive_types


class Market(MarketBase):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._exchange = 'some_exchange_1'
        self._market_type = "spot"
        self._base_ticker = "BTC"
        self._quote_ticker = "USDT"
        self._price_precision = 2
        self._lot_precision = 4

    def fetch_price(self, symbol: Optional[str] = None) -> Decimal: pass
    def fetch_ohlc(self, timeframe: str, symbol: Optional[str] = None, columns: Optional[Iterable[str]] = None,
                   index_by_date: bool = True, limit: int = 500) -> Iterable[dict]: pass


class AnotherMarket(MarketBase):
    def fetch_price(self, symbol: Optional[str] = None) -> Decimal: pass
    def fetch_ohlc(self, timeframe: str, symbol: Optional[str] = None, columns: Optional[Iterable[str]] = None,
                   index_by_date: bool = True, limit: int = 500) -> Iterable[dict]: pass


class TestMarketBase(unittest.TestCase):
    def setUp(self):
        # Clean up all singleton instances as they may exist from previous tests
        MarketBase._singleton_instances = {}

        # Create Market object
        self.symbol = "BTCUSDT"
        self.market = Market(self.symbol)

        # Create mocked Ohlc object
        self.timeframe = '1h'
        self.ohlc = Mock(timeframe=self.timeframe, need_update=True, updated=0)

    def test_singleton(self):
        # Test singleton same credentials
        instance1 = Market(self.symbol)
        instance2 = Market(self.symbol)
        self.assertIs(instance1, instance2)
        self.assertEqual(1, len(MarketBase._singleton_instances))
        self.assertEqual(1, instance1.n_instances)
        self.assertEqual(1, instance2.n_instances)

        # Test singleton different symbol
        instance3 = Market('ETHUSDT')
        self.assertIs(instance1, instance2)
        self.assertIsNot(instance1, instance3)
        self.assertIsNot(instance2, instance3)
        self.assertEqual(2, len(MarketBase._singleton_instances))
        self.assertEqual(2, instance1.n_instances)
        self.assertEqual(2, instance2.n_instances)
        self.assertEqual(2, instance3.n_instances)

        # Test tbe singleton logic works right in case of another child
        instance4 = AnotherMarket(self.symbol)
        self.assertIs(instance1, instance2)
        self.assertIsNot(instance3, instance1)
        self.assertIsNot(instance3, instance2)
        self.assertIsNot(instance4, instance1)
        self.assertIsNot(instance4, instance2)
        self.assertIsNot(instance4, instance3)
        self.assertEqual(3, len(MarketBase._singleton_instances))
        self.assertEqual(2, instance1.n_instances)
        self.assertEqual(2, instance2.n_instances)
        self.assertEqual(2, instance3.n_instances)
        self.assertEqual(1, instance4.n_instances)

    def test_initialization(self):
        self.assertGreater(self.market.initialized, 0)
        self.assertEqual(self.market.base_ticker, "BTC")
        self.assertEqual(self.market.quote_ticker, "USDT")
        self.assertEqual(self.market.exchange, 'some_exchange_1')
        self.assertEqual(self.market.market_type, 'spot')
        self.assertEqual(self.market.symbol, "BTCUSDT")
        self.assertEqual(self.market.price, Decimal(0))
        self.assertEqual(self.market.price_time, 0)
        self.assertIsInstance(self.market._ohlc, dict)
        self.assertDictEqual(self.market._ohlc, {})
        self.assertIsInstance(self.market.price_precision, int)
        self.assertEqual(self.market.price_precision, 2)
        self.assertIsInstance(self.market.price_step, Decimal)
        self.assertEqual(self.market.price_step, Decimal(0))
        self.assertIsInstance(self.market.lot_precision, int)
        self.assertEqual(self.market.lot_precision, 4)
        self.assertIsInstance(self.market.lot_step, Decimal)
        self.assertEqual(self.market.lot_step, Decimal(0))
        self.assertIsInstance(self.market.min_order_base, Decimal)
        self.assertEqual(self.market.min_order_base, Decimal(0))
        self.assertIsInstance(self.market.min_order_quote, Decimal)
        self.assertEqual(self.market.min_order_quote, Decimal(0))
        self.assertDictEqual(self.market.provider_data, {})

        # Try to get the market using get_market
        market = MarketBase.get_market(self.market.exchange, self.market.market_type, self.market.symbol)
        self.assertIsNotNone(market)
        self.assertIs(self.market, market)  # Should return the same instance

        # Test market initialization without child __init__()
        another = AnotherMarket(self.symbol)
        self.assertEqual(another.base_ticker, "")
        self.assertEqual(another.quote_ticker, "")
        self.assertEqual(another.exchange, '')
        self.assertEqual(another.market_type, '')
        self.assertIsNot(another, market)

        # Test .get_market() not found not existing market
        market = MarketBase.get_market("different_exchange", self.market.market_type, self.market.symbol)
        self.assertIsNone(market)

        # Try to get the market using lowercase symbol
        market = MarketBase.get_market(self.market.exchange, self.market.market_type, "btcusdt")
        self.assertIsNotNone(market)
        self.assertIs(self.market, market)

        # Try to get instance via another child class
        market = AnotherMarket.get_market(self.market.exchange, self.market.market_type, self.market.symbol)
        self.assertIsNotNone(market)
        self.assertIs(self.market, market)

        # Test access to timeframes
        self.assertEqual(0, len(self.market))
        for tf in TIMEFRAMES:
            self.assertNotIn(tf, self.market)
            self.assertNotIn(tf, self.market._ohlc)
        self.assertEqual(0, len(self.market))  # check the length after iterating

        # Test .min_order_base dynamic property recalculation
        self.market._min_order_quote = Decimal(5)
        self.market._price = Decimal(100000)
        self.market._lot_precision = 5
        self.assertEqual(self.market.min_order_quote, Decimal(5))
        self.assertEqual(self.market.price, Decimal(100000))
        self.assertEqual(self.market.min_order_base, Decimal(5)/Decimal(100000))
        self.assertEqual(self.market._min_order_base, Decimal(5)/Decimal(100000))
        self.market._min_order_base = Decimal(0)
        self.assertEqual(self.market._min_order_base, 0)
        self.assertEqual(self.market.min_order_base, Decimal(5)/Decimal(100000))

    # @patch('amarket.interfaces.i_ohlc.Ohlc')
    def test_access_timeframes(self):
        with self.assertRaises(ValueError):
            self.market['some_not_timeframe_key'] = self.ohlc

        with self.assertRaises(KeyError):
            _ = self.market['some_not_timeframe_key']

        with self.assertRaises(KeyError):
            _ = self.market[self.timeframe]

        with self.assertRaises(ValueError):
            # noinspection PyTypeChecker
            self.market[self.timeframe] = None

        with self.assertRaises(ValueError):
            self.market['1s'] = self.ohlc  # check wrong timeframe assignment

        self.market._ohlc[self.timeframe] = self.ohlc  # direct OHLC data assignment, for testing purposes only!

        # Check the instance after adding valid OHLC data
        self.assertIn(self.timeframe, self.market)
        self.assertIn(self.timeframe, self.market._ohlc)
        self.assertEqual(1, len(self.market))
        self.assertEqual(1, len(self.market._ohlc))
        self.assertIsInstance(self.market._ohlc, dict)

        # Test iterations over existing timeframes
        cnt = 0
        for _ in self.market:
            self.assertIn(_, TIMEFRAMES)
            cnt += 1
        self.assertEqual(cnt, len(self.market))
        self.assertEqual(cnt, len(self.market._ohlc))

        # Check the OHLC storage is different for another market instance
        another_market = AnotherMarket(self.symbol)
        self.assertIsInstance(self.market._ohlc, dict)
        self.assertNotEqual(self.market._ohlc, another_market._ohlc)
        self.assertIsNot(self.market._ohlc, another_market._ohlc)
        self.assertIsNot(self.market, another_market)
        self.assertEqual(1, len(self.market))
        self.assertEqual(0, len(another_market))

        # Check the instances after adding new valid OHLC data to another market
        another_market._ohlc[self.timeframe] = Mock(timeframe=self.timeframe)
        self.assertNotEqual(self.market._ohlc, another_market._ohlc)
        self.assertIsNot(self.market._ohlc, another_market._ohlc)
        self.assertIsNot(self.market, another_market)
        self.assertEqual(1, len(self.market))
        self.assertEqual(1, len(another_market))

    def test_update(self):
        self.market._ohlc[self.timeframe] = self.ohlc
        self.market._ohlc["1d"] = Mock(timeframe='1d', need_update=True, updated=0)
        self.assertTrue(self.market[self.timeframe].need_update)
        self.assertTrue(self.market["1d"].need_update)

        # Test force update
        self.market[self.timeframe].need_update = False
        self.market["1d"].need_update = False
        self.market.update(force=[self.timeframe, "1d"])
        self.assertTrue(self.market._ohlc[self.timeframe].need_update)

    def test_update_min_fetch_interval(self):
        """Test the min_fetch_interval parameter in update method"""
        # Mock the time and fetch methods
        self.market._ohlc[self.timeframe] = self.ohlc

        # Test that min_fetch_interval defaults to 0 when not provided
        self.market.fetch_price = Mock(return_value=Decimal('1000.0'))
        self.market.fetch_ohlc = Mock()

        # Generate the function arguments to test
        values = (None, 0, 0.5, True, False, -1000, "invalid", 1000, "1000", 400, 1)
        arguments = ([{'price': True}] +
                     [{'price': True, 'min_fetch_interval': _} for _ in values] +
                     [{'force': [self.timeframe]}] +
                     [{'force': [self.timeframe], 'min_fetch_interval': _} for _ in values])

        # Test with initial price_time and ohlc.updated (both are zero) - always fetch
        for kwargs in arguments:
            self.market.update(**kwargs)
            self.market.fetch_price.assert_called_once()
            self.market.fetch_ohlc.assert_called_once()
            self.market.fetch_price.reset_mock()
            self.market.fetch_ohlc.reset_mock()
            self.market._price_time, self.market._ohlc_time = 0, 0

        # Set price_time and ohlc.updated to current time minus 10 seconds and test that update happens - always fetch
        current_time = int(time() * 1000) - 10000
        self.market._price_time, self.market._ohlc_time = current_time, current_time
        for kwargs in arguments:
            self.market.update(**kwargs)
            self.market.fetch_price.assert_called_once()
            self.market.fetch_ohlc.assert_called_once()
            self.market.fetch_price.reset_mock()
            self.market.fetch_ohlc.reset_mock()
            self.market._price_time, self.market._ohlc_time = current_time, current_time

        # Set price_time and ohlc.updated to current time and test that update happens only when:
        # min_fetch_interval is explicitly disabled (None or 0)
        current_time = int(time() * 1000)
        self.market._price_time, self.market._ohlc_time = current_time, current_time
        for kwargs in arguments:
            self.market.update(**kwargs)
            if 'min_fetch_interval' in kwargs and kwargs['min_fetch_interval'] in (None, 0):
                self.market.fetch_price.assert_called_once()
                self.market.fetch_ohlc.assert_called_once()
                self.market.fetch_price.reset_mock()
                self.market.fetch_ohlc.reset_mock()
                self.market._price_time, self.market._ohlc_time = current_time, current_time
            else:
                self.market.fetch_price.assert_not_called()
                self.market.fetch_ohlc.assert_not_called()

        # Set price_time and ohlc.updated to current time - 0.5 seconds and test that update happens only when:
        # min_fetch_interval is integer and explicitly less than 500 OR explicitly disabled (None or 0)
        current_time = int(time() * 1000) - 500
        self.market._price_time, self.market._ohlc_time = current_time, current_time
        for kwargs in arguments:
            self.market.update(**kwargs)
            in_kwargs = 'min_fetch_interval' in kwargs
            disabled = in_kwargs and kwargs['min_fetch_interval'] in (None, 0)
            is_bool = in_kwargs and kwargs['min_fetch_interval'] in (False, True)
            is_int = in_kwargs and not is_bool and isinstance(kwargs['min_fetch_interval'], int)
            is_low_positive_int = is_int and 500 > kwargs['min_fetch_interval'] > 0
            if is_low_positive_int or disabled:
                self.market.fetch_price.assert_called_once()
                self.market.fetch_ohlc.assert_called_once()
                self.market.fetch_price.reset_mock()
                self.market.fetch_ohlc.reset_mock()
                self.market._price_time, self.market._ohlc_time = current_time, current_time
            else:
                self.market.fetch_price.assert_not_called()
                self.market.fetch_ohlc.assert_not_called()

    def test_validate(self):
        # Test Market passes validation with correct properties
        result = self.market.validate()
        self.assertTrue(result["success"])
        self.assertEqual(len(result["errors"]), 0)

        # Test validation fails with invalid properties
        invalid_props = {'exchange': '', 'market_type': "", 'symbol': "", 'base_ticker': "", 'quote_ticker': "",
                         'price_precision': -1, 'lot_precision': -1}
        for prop, value in invalid_props.items():
            init_value = getattr(self.market, f"_{prop}")
            setattr(self.market, f"_{prop}", value)
            result = self.market.validate()
            self.assertFalse(result["success"])
            self.assertIn(f"Invalid {prop}", result["errors"][0])
            self.assertEqual(len(result['errors']), 1)
            setattr(self.market, f"_{prop}", init_value)  # restore

    def test_to_dict(self):
        expected_dict = {
            "exchange": 'some_exchange_1', "market_type": "spot", "symbol": "BTCUSDT",
            "base_ticker": "BTC", "quote_ticker": "USDT", "price": Decimal(0), "price_time": 0, "price_precision": 2,
            "price_step": Decimal(0), "lot_precision": 4, "lot_step": Decimal(0), "min_order_base": Decimal(0),
            "min_order_quote": Decimal(0), "provider_data": {}, "timeframes": []}

        result = self.market.to_dict()
        self.assertEqual(result, expected_dict)

        # Test with modified values
        self.market._price = Decimal('10000')
        self.market._price_time = int(time() * 1000)
        self.market._ohlc['1h'] = Mock(timeframe='1h', need_update=True, updated=0)

        expected_dict['price'] = self.market._price
        expected_dict['price_time'] = self.market._price_time
        expected_dict['timeframes'] = ['1h']

        result = self.market.to_dict()
        self.assertEqual(result, expected_dict)
        check_primitive_types(result)

    def test_round_lot(self):
        """Test the round_lot method"""
        # Test with different lot steps and values using default rounding (ROUND_HALF_EVEN)
        test_cases = [
            (Decimal('0.0001'), Decimal('0.12345'), Decimal('0.1234')),  # Normal case, precision=4
            (Decimal('0.0001'), Decimal('0.12349'), Decimal('0.1235')),  # Round up, precision=4
            (Decimal('0.0001'), Decimal('0.12341'), Decimal('0.1234')),  # Round down, precision=4
            (Decimal('0.0001'), Decimal('0.0'), Decimal('0.0')),        # Zero value, precision=4
            (Decimal('0.0001'), Decimal('1.0'), Decimal('1.0')),        # Whole number, precision=4
            (Decimal('0.000001'), Decimal('0.1234567'), Decimal('0.123457')),  # Higher precision=6, rounds to even
            (Decimal('1'), Decimal('1.7'), Decimal('2.0')),         # Zero precision should round to nearest integer
        ]

        for lot_step, value, expected in test_cases:
            self.market._lot_step = lot_step
            result = self.market.round_lot(value)
            self.assertEqual(result, expected, f"Failed for lot_step={lot_step}, value={value}")

        # Test different rounding modes with midpoint values
        midpoint_value = Decimal('0.12345')  # Will round to 0.1234 or 0.1235 depending on rounding mode
        self.market._lot_step = Decimal('0.0001')

        test_rounding_cases = [
            (None, 'ROUND_HALF_EVEN', Decimal('0.1234')),  # Default: rounds to even (4)
            (ROUND_UP, 'ROUND_UP', Decimal('0.1235')),
            (ROUND_DOWN, 'ROUND_DOWN', Decimal('0.1234')),
            (ROUND_HALF_UP, 'ROUND_HALF_UP', Decimal('0.1235')),  # Rounds away from zero on tie
            (ROUND_HALF_DOWN, 'ROUND_HALF_DOWN', Decimal('0.1234')),  # Rounds toward zero on tie
            (ROUND_CEILING, 'ROUND_CEILING', Decimal('0.1235')),
            (ROUND_FLOOR, 'ROUND_FLOOR', Decimal('0.1234')),
        ]

        for rounding, description, expected in test_rounding_cases:
            result = self.market.round_lot(midpoint_value, rounding=rounding)
            self.assertEqual(result, expected, f"Failed for rounding={description}, value={midpoint_value}")

    def test_round_price(self):
        """Test the round_price method"""
        # Test with different price steps and values using default rounding (ROUND_HALF_EVEN)
        test_cases = [
            (Decimal('0.01'), Decimal('100.1234'), Decimal('100.12')),  # Normal case, precision=2
            (Decimal('0.01'), Decimal('100.1235'), Decimal('100.12')),  # Round down at midpoint, precision=2
            (Decimal('0.01'), Decimal('100.1285'), Decimal('100.13')),  # Round up, precision=2
            (Decimal('0.01'), Decimal('100.0'), Decimal('100.00')),     # Zero value, precision=2
            (Decimal('0.01'), Decimal('101.0'), Decimal('101.00')),     # Whole number, precision=2
            (Decimal('0.0001'), Decimal('99.99995'), Decimal('100.0000')),  # Higher precision=4, rounds to even
            (Decimal('1'), Decimal('99.6'), Decimal('100.0')),        # Zero precision should round to nearest integer
        ]

        for price_step, value, expected in test_cases:
            self.market._price_step = price_step
            result = self.market.round_price(value)
            self.assertEqual(result, expected, f"Failed for price_step={price_step}, value={value}")

        # Test different rounding modes with midpoint values
        midpoint_value = Decimal('100.125')  # Will round to 100.12 or 100.13 depending on rounding mode
        self.market._price_step = Decimal('0.01')

        test_rounding_cases = [
            (None, 'ROUND_HALF_EVEN', Decimal('100.12')),  # Default: rounds to even (2)
            (ROUND_UP, 'ROUND_UP', Decimal('100.13')),
            (ROUND_DOWN, 'ROUND_DOWN', Decimal('100.12')),
            (ROUND_HALF_UP, 'ROUND_HALF_UP', Decimal('100.13')),  # Rounds away from zero on tie
            (ROUND_HALF_DOWN, 'ROUND_HALF_DOWN', Decimal('100.12')),  # Rounds toward zero on tie
            (ROUND_CEILING, 'ROUND_CEILING', Decimal('100.13')),
            (ROUND_FLOOR, 'ROUND_FLOOR', Decimal('100.12')),
        ]

        for rounding, description, expected in test_rounding_cases:
            result = self.market.round_price(midpoint_value, rounding=rounding)
            self.assertEqual(result, expected, f"Failed for rounding={description}, value={midpoint_value}")


if __name__ == "__main__":
    unittest.main()



import unittest
from datetime import datetime, timezone
from decimal import Decimal
from time import time

from amarket import SymbolAccountError, SymbolAccountMixin, SymbolError, SymbolMixin


class SymbolMixinTests(unittest.TestCase):
    def setUp(self):
        self.mixin = SymbolMixin(base_ticker="btc", quote_ticker="USD")

    def test_initialization(self):
        self.assertEqual(self.mixin.base_ticker, "BTC")
        self.assertEqual(self.mixin.quote_ticker, "USD")
        self.assertEqual(self.mixin.symbol, "BTCUSD")
        with self.assertRaises(AttributeError):
            self.mixin.base_ticker = 'ETH'  # noqa
        with self.assertRaises(AttributeError):
            self.mixin.quote_ticker = 'USDT'  # noqa
        with self.assertRaises(AttributeError):
            self.mixin.symbol = 'ETHUSD'  # noqa

        mixin = SymbolMixin()
        self.assertEqual(mixin.base_ticker, SymbolMixin.CURRENCY_BASE.upper())
        self.assertEqual(mixin.quote_ticker, SymbolMixin.CURRENCY_QUOTE.upper())
        self.assertEqual(mixin.symbol, "BASEQUOTE")

    def test_init_with_empty_tickers(self):
        with self.assertRaises(SymbolError):
            SymbolMixin(base_ticker="", quote_ticker="")
        with self.assertRaises(SymbolError):
            SymbolMixin(base_ticker="", quote_ticker="USDT")
        with self.assertRaises(SymbolError):
            SymbolMixin(base_ticker="btc", quote_ticker="")

    def test_init_with_long_tickers(self):
        with self.assertRaises(SymbolError):
            SymbolMixin(base_ticker="ETHETHETHETH", quote_ticker="USD")
        with self.assertRaises(SymbolError):
            SymbolMixin(base_ticker="BTC", quote_ticker="usdusdusdusd")
        with self.assertRaises(SymbolError):
            SymbolMixin(base_ticker="ETHETHEThETH", quote_ticker="USD", ticker_max_length=11)
        mixin = SymbolMixin(base_ticker="ETHETHEThETH", quote_ticker="USD", ticker_max_length=12)
        self.assertEqual(mixin.base_ticker, "ETHETHETHETH")
        self.assertEqual(mixin.quote_ticker, "USD")
        self.assertEqual(mixin.symbol, "ETHETHETHETHUSD")

    def test_price(self):
        self.assertEqual(self.mixin.price, 0.0)
        for value, expected in [(100.0, 100.0), (Decimal('99.99'), 99.99)]:
            timestamp = time()
            self.mixin.price = value
            self.assertEqual(self.mixin.price, expected)
            self.assertGreaterEqual(self.mixin.price_timestamp, round(timestamp))
            self.assertGreaterEqual(self.mixin.price_timestamp_ms, round(timestamp * 1000))

        with self.assertRaises(SymbolError):
            self.mixin.price = -10.0

        with self.assertRaises(SymbolError):
            self.mixin.price = 'invalid'

    def test_price_timestamp(self):
        timestamp = time()
        for value, expected1, expected2, expected3 in [
            (timestamp, round(timestamp), round(timestamp*1000), datetime.fromtimestamp(timestamp, tz=timezone.utc)),
            (1234567890.123456, 1234567890, 1234567890123, datetime.fromtimestamp(1234567890.123456, tz=timezone.utc)),
        ]:
            self.mixin.price_timestamp = value
            self.assertEqual(self.mixin.price_timestamp, expected1)
            self.assertEqual(self.mixin.price_timestamp_ms, expected2)
            self.assertEqual(self.mixin.price_datetime, expected3)

        with self.assertRaises(SymbolError):
            self.mixin.price_timestamp = -10.0

        with self.assertRaises(SymbolError):
            self.mixin.price_timestamp = 'invalid'


class SymbolBalanceMixinTests(unittest.TestCase):
    def setUp(self):
        self.mixin = SymbolAccountMixin(base_quantity=10, quote_quantity=100)

    def test_initialization(self):
        self.assertEqual(self.mixin.base_quantity, 10)
        self.assertEqual(self.mixin.quote_quantity, 100)

        mixin = SymbolAccountMixin('useless_arg1', 'useless_arg2',
                                   base_quantity=0, quote_quantity=0, useless_parameter=True)
        self.assertEqual(mixin.base_quantity, 0)
        self.assertEqual(mixin.quote_quantity, 0)

        with self.assertRaises(SymbolAccountError):
            SymbolAccountMixin(base_quantity=-10, quote_quantity=100)

        with self.assertRaises(SymbolAccountError):
            SymbolAccountMixin(base_quantity=10, quote_quantity=-100)


if __name__ == '__main__':
    unittest.main()



import unittest

import pandas as pd
from amarket.ohlc import cooldown, ohlc_cleanup
from amarket.test_data import DATAFRAMES


class TestOhlcFunctions(unittest.TestCase):
    def setUp(self):
        self.dataframes = DATAFRAMES

    def test_cooldown(self):
        # Test cases dictionary where key is input, value is a list of test cases
        data = {
            (): [  # test empty series
                (0, [])
            ],

            (0, 0, 0): [  # no signals case
                (1, (0, 0, 0)),
                (2, (0, 0, 0)),
            ],

            (1, 0, 0): [  # single signal
                (1, (1, 0, 0)),
                (2, (1, 0, 0)),
            ],

            (1, 1, 0): [  # multiple signals within cooldown
                (1, (1, 0, 0)),
                (2, (1, 0, 0)),
            ],

            (1, 0, 1): [  # multiple signals outside cooldown
                (1, (1, 0, 1)),
                (2, (1, 0, 0))
            ],

            (1, 0, 0, 0, 0, 1): [  # large cooldown
                (4, (1, 0, 0, 0, 0, 1))
            ],

            (1, 0, 1, 1, 0, 1): [
                (1, [1, 0, 1, 0, 0, 1]),
                (2, [1, 0, 0, 1, 0, 0]),
                (3, [1, 0, 0, 0, 0, 1]),
            ],

            (1, 1, 1, 1, 1, 1): [
                (1, [1, 0, 1, 0, 1, 0]),
                (2, [1, 0, 0, 1, 0, 0]),
            ],

            (1, 1, 1, -1, -1, -1, 0, 0, 0): [
                (2, [1, 0, 0, -1, 0, 0, 0, 0, 0]),
            ],

            (-1, 0, 1, -1, 0, 1, -1, 0, 1): [
                (2, [-1, 0, 1, -1, 0, 1, -1, 0, 1]),
                (3, [-1, 0, 1, 0, 0, 0, -1, 0, 1]),
            ],

            (1, -1, 1, -1, 1, -1, 1, -1, 1): [
                (5, [1, -1, 0, 0, 0, 0, 1, -1, 0])
            ],
        }

        for signal, test_cases in data.items():
            # First, test periods lower or equal to 0
            for period in range(-3, 1):
                self.assertEqual(pd.Series(signal).tolist(), cooldown(pd.Series(signal), period=period).tolist(),
                                 msg=f'StrategyBase.cooldown({signal}, period={period}) works wrong!')

            # Then, test with payload
            for period, expected in test_cases:
                inp = pd.Series(signal).astype(int)
                exp = pd.Series(expected).astype(int)
                self.assertEqual(exp.tolist(), cooldown(inp, period=period).tolist(),
                                 msg=f'StrategyBase.cooldown({signal}, period={period}) works wrong!')
                self.assertIsNone(pd.testing.assert_series_equal(exp, cooldown(inp, period=period)))

    def test_ohlc_cleanup(self):
        symbol, timeframe = 'BTCUSDT', '1d'
        df_key = f"{symbol}_{timeframe}"
        df = self.dataframes[df_key].copy()
        df.loc[len(df)] = [pd.to_datetime('2022-12-30', utc=True), 1, 2, 3, 4, 5]
        df.loc[len(df)] = [pd.to_datetime('2022-12-30', utc=True), 1, 2, 3, 4, 5]  # duplicate row
        df.loc[len(df)] = [pd.to_datetime('2021-01-01', utc=True), 0, 0, 0, 0, 0]  # zero volume row
        dataframe, _ = ohlc_cleanup(df, timeframe, date_index=True)

        deltas = dataframe.index.diff().dropna()
        self.assertEqual(len(dataframe), len(self.dataframes[df_key]) + 2)
        self.assertTrue(dataframe.index.is_monotonic_increasing, msg="Chronological order failed!")
        self.assertEqual(pd.Timedelta(timeframe), dataframe.index[1] - dataframe.index[0])
        self.assertTrue((deltas == (dataframe.index[1] - dataframe.index[0])).all())
        self.assertTrue((deltas == pd.Timedelta(timeframe)).all())
        self.assertTrue((dataframe.loc['2022-12-31'] == [1, 2, 3, 4, 5]).all(), msg="Backfill failed!")
        self.assertLessEqual(dataframe.duplicated().sum(), 1, msg="There is duplicated rows!")


if __name__ == '__main__':
    unittest.main()



import unittest

from amarket import AssetSet, AssetSetError


class AssetSetTests(unittest.TestCase):
    def setUp(self):
        self.tickers = ['AAPL', 'GOOG']
        self.amounts = [10, 20]
        self.asset_set = AssetSet(self.tickers, self.amounts)

    def test_init_with_valid_data(self):
        tickers = ['AAPL', 'GOOG', 'MSFT']
        amounts = [10, 20, 30]
        asset_set = AssetSet(tickers, amounts)
        self.assertEqual(len(asset_set), 3)
        self.assertEqual(len(asset_set.tickers), 3)
        self.assertEqual(len(asset_set.amounts), 3)
        self.assertEqual(len(asset_set.assets), 3)
        self.assertEqual(asset_set.n, 3)
        self.assertEqual(asset_set['AAPL'], 10)
        self.assertEqual(asset_set['GOOG'], 20)
        self.assertEqual(asset_set['MSFT'], 30)

        asset_set = AssetSet([], [])
        self.assertEqual(len(asset_set), 0)
        self.assertEqual(asset_set.n, 0)

    def test_init_with_invalid_tickers(self):
        tickers = ['AAPL', 'GOOG', 123]
        amounts = [10, 20, 30]
        with self.assertRaises(AssetSetError):
            AssetSet(tickers, amounts)

    def test_init_with_invalid_amounts(self):
        tickers = ['AAPL', 'GOOG', 'MSFT']
        amounts = [10, '20', 30]
        with self.assertRaises(AssetSetError):
            AssetSet(tickers, amounts)

    def test_init_with_duplicate_tickers(self):
        tickers = ['AAPL', 'GOOG', 'AAPL']
        amounts = [10, 20, 30]
        with self.assertRaises(AssetSetError):
            AssetSet(tickers, amounts)

    def test_init_with_mismatched_lengths(self):
        tickers = ['AAPL', 'GOOG', 'MSFT']
        amounts = [10, 20]
        with self.assertRaises(AssetSetError):
            AssetSet(tickers, amounts)

    def test_setitem(self):
        asset_set = AssetSet(['AAPL'], [10])
        asset_set['GOOG'] = 20
        self.assertEqual(len(asset_set), 2)
        self.assertEqual(asset_set.n, 2)
        self.assertEqual(asset_set['AAPL'], 10)
        self.assertEqual(asset_set['GOOG'], 20)

    def test_delitem(self):
        del self.asset_set['AAPL']
        self.assertEqual(len(self.asset_set), 1)
        self.assertEqual(self.asset_set.n, 1)
        with self.assertRaises(KeyError):
            self.asset_set['AAPL']  # noqa

    def test_getitem(self):
        # with int key
        self.assertEqual(self.asset_set[0], 10)
        self.assertEqual(self.asset_set[1], 20)

        # with str key
        self.assertEqual(self.asset_set['AAPL'], 10)
        self.assertEqual(self.asset_set['GOOG'], 20)

        # with_invalid_key_type
        with self.assertRaises(TypeError):
            self.asset_set[1.5]  # noqa

        # with non-existing ticker
        with self.assertRaises(KeyError):
            self.asset_set['MSFT']  # noqa

    def test_iter(self):
        items = list(self.asset_set)
        self.assertEqual(items, self.tickers)

        items = list(dict(zip(self.tickers, self.amounts)).items())
        self.assertEqual(items, list(self.asset_set.items()))

    def test_next(self):
        iterator = iter(self.asset_set)
        self.assertEqual(next(iterator), 'AAPL')
        self.assertEqual(next(iterator), 'GOOG')
        with self.assertRaises(StopIteration):
            next(iterator)

    def test_comparison(self):
        asset_set = AssetSet(self.tickers, [20, 30])
        self.assertNotEqual(self.asset_set, asset_set)

        asset_set = AssetSet(['AAPL', 'MSFT'], [10, 20])
        self.assertNotEqual(self.asset_set, asset_set)

        asset_set = AssetSet(self.tickers, self.amounts)
        self.assertEqual(self.asset_set, asset_set)

    def test_sub(self):
        asset_set1 = AssetSet(['AAPL', 'GOOG'], [10, 20])
        asset_set2 = AssetSet(['GOOG', 'MSFT'], [15, 25])
        result = asset_set1 - asset_set2
        self.assertEqual(len(result), 3)
        self.assertEqual(result.n, 3)
        self.assertEqual(result['AAPL'], 10)
        self.assertEqual(result['GOOG'], 5)
        self.assertEqual(result['MSFT'], -25)

        asset_set1 = AssetSet(['AAPL', 'GOOG', 'MSFT'], [10, 20, 30])
        result = asset_set1 - asset_set2
        self.assertEqual(len(result), 3)
        self.assertEqual(result.n, 3)
        self.assertEqual(result['AAPL'], 10)
        self.assertEqual(result['GOOG'], 5)
        self.assertEqual(result['MSFT'], 5)

    def test_value(self):
        prices = [100, 200]
        self.assertEqual(self.asset_set.value(prices), 5000)

        asset_set = AssetSet(['AAPL', 'GOOG', 'MSFT'], [10, 20, 30])
        prices = [100, 200, 300]
        expected_value = sum([a * p for a, p in zip(asset_set.amounts, prices)])
        self.assertEqual(asset_set.value(prices), expected_value)

        with self.assertRaises(ValueError):
            asset_set.value([100, 200])

    def test_weights(self):
        prices = [100.5, 38.12]
        value = sum([price * amount for price, amount in zip(prices, self.asset_set.amounts)])
        weights = [price * amount / value for price, amount in zip(prices, self.asset_set.amounts)]
        self.assertEqual(self.asset_set.weights(prices), weights)

    def test_sort(self):
        asset_set = AssetSet(['GOOG', 'AAPL', 'MSFT'], [20, 10, 30])
        self.assertEqual(asset_set.tickers, ['AAPL', 'GOOG', 'MSFT'])
        self.assertEqual(asset_set.amounts, [10, 20, 30])


if __name__ == '__main__':
    unittest.main()



import unittest
from datetime import datetime, timezone
from decimal import Decimal
from zoneinfo import ZoneInfo

from amarket.functions import (
    check_interval_overlap,
    convert_to_timestamp,
    find_nearest_idx,
    fmt,
    format_number,
    is_sorted,
    is_timestamp_ms,
)
from amarket.functions.utils import s_fmt


class TestFunctions(unittest.TestCase):
    def test_is_sorted(self):
        arr = [91233.01, 93212.22, 94183, 97145.19, 100_000.11, 100_000.11]
        self.assertTrue(is_sorted(arr))
        self.assertFalse(is_sorted(arr, reverse_order=True))
        self.assertTrue(is_sorted(reversed(arr), reverse_order=True))
        self.assertFalse(is_sorted(arr, unique=True))
        self.assertTrue(is_sorted(arr[:-1], unique=True))
        self.assertFalse(is_sorted(arr, reverse_order=True, unique=True))
        self.assertTrue(is_sorted(reversed(arr[:-1]), reverse_order=True, unique=True))

    def test_find_nearest_idx(self):
        arr = [91233.01, 93212.22, 94183, 97145.19, 100_000.11]
        # Format of test tuple: (input_value, closest_index, left_index, right_index)
        test = [(95000, (2, 2, 3)), (950009, (None, None, None)), (909, (None, None, None)),
                (94001.21, (2, 1, 2)), (96000, (3, 2, 3)), (99999.99, (4, 3, 4)),
                (94183, (2, 1, 3)), (100_000.11, (4, 3, 4)), (91233.01, (0, 0, 1))]
        for val, exp in test:
            self.assertEqual(find_nearest_idx(arr, val), exp)

    def test_check_interval_overlap(self):
        data = [((0, 5), (3, 4), True), ((0, 5), (3, 7), True), ((0, 5), (6, 7), False),
                ((0, 5), (-1, 6), True), ((0, 5), (-1, 3), True), ((0, 5), (7, 8), False),
                ((0, 5), (-4, 0), True), ((0, 5), (-4, -0.2), False)]
        for interval1, interval2, expected in data:
            self.assertEqual(check_interval_overlap(interval1, interval2), expected)

    def test_convert_to_timestamp(self):
        expected = 1609459200  # 2021-01-01T00:00:00 in UTC

        # Test with Unix timestamp in seconds
        self.assertEqual(convert_to_timestamp(expected), expected)

        # Test with Unix timestamp in milliseconds (should convert to seconds)
        self.assertEqual(convert_to_timestamp(expected * 1000), expected)

        # Test with ISO format datetime string
        self.assertEqual(convert_to_timestamp("2021-01-01T00:00:00"), expected)

        # Test with ISO format datetime string with timezone info
        self.assertEqual(convert_to_timestamp("2021-01-01T00:00:00+00:00"), expected)
        self.assertEqual(convert_to_timestamp("2021-01-01T00:00:00Z"), expected)

        # Test with GMT+3 timezone
        self.assertEqual(convert_to_timestamp("2021-01-01T03:00:00+03:00"), expected)

        # Test with invalid input
        self.assertIsNone(convert_to_timestamp("invalid"))

        # Test milliseconds detection using decimal digits logic
        current_ts = int(datetime.now().timestamp())
        future_ms = (current_ts + 1000) * 1000  # a milliseconds timestamp
        self.assertEqual(convert_to_timestamp(future_ms), current_ts + 1000)

        # Test with very large value that should be treated as milliseconds
        large_value = 9999999999999  # it should be treated as ms
        self.assertEqual(convert_to_timestamp(large_value), round(large_value / 1000))

    def test_fmt(self):
        # Fixed timestamp for testing (2021-01-01 12:00:00 UTC)
        fixed_ts = 1609459200
        dt_obj = datetime(2021, 1, 1, 0, 0, 0, tzinfo=timezone.utc)
        expected = "2021-01-01 00:00:00"

        # Test with datetime object
        self.assertEqual(expected, fmt(dt_obj))

        # Test with timestamp
        expected_format = datetime.fromtimestamp(fixed_ts, tz=timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
        self.assertEqual(expected_format, fmt(fixed_ts))

        # Test with custom format
        custom_format = "%d/%m/%Y"
        self.assertEqual("01/01/2021", fmt(dt_obj, fmt=custom_format))

        # Test with different timezone - UTC using ZoneInfo
        utc_tz = ZoneInfo('UTC')
        dt_utc = datetime(2021, 1, 1, 0, 0, 0, tzinfo=utc_tz)
        self.assertEqual(expected, fmt(dt_utc))
        self.assertEqual(expected, fmt(fixed_ts, tz=timezone.utc))
        self.assertEqual(expected, fmt(fixed_ts, tz=utc_tz))
        self.assertEqual(expected, fmt(fixed_ts, tz='UTC'))

        # Test with different timezone - Europe/Moscow using ZoneInfo
        msk_tz = ZoneInfo('Europe/Moscow')
        dt_msk = datetime(2021, 1, 1, 3, 0, 0, tzinfo=msk_tz)
        self.assertEqual(expected, fmt(dt_msk))
        self.assertEqual(expected, fmt(dt_msk, tz=utc_tz))
        self.assertEqual(expected, fmt(dt_msk, tz='UTC'))

        # Note: The timestamp will be converted to Moscow time
        moscow_time = datetime.fromtimestamp(fixed_ts, tz=msk_tz).strftime('%Y-%m-%d %H:%M:%S')
        self.assertEqual(moscow_time, fmt(fixed_ts, tz=msk_tz))

        # Test that fmt() correctly handles timestamps in milliseconds
        # Get current timestamp in seconds and milliseconds
        current_ts = int(datetime.now().timestamp())
        current_ms = current_ts * 1000
        expected_format = datetime.fromtimestamp(current_ts, tz=timezone.utc).strftime('%Y-%m-%d %H:%M:%S')

        # Test with milliseconds timestamp - should auto-detect and convert to seconds
        self.assertEqual(expected_format, fmt(current_ms))

        # Test with future timestamps in milliseconds
        future_sec = current_ts + 1000
        future_ms = future_sec * 1000
        expected_future = datetime.fromtimestamp(future_sec, tz=timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
        self.assertEqual(expected_future, fmt(future_ms))

        # Test with very large value that should be treated as milliseconds
        large_value = 9999999999999  # This should be treated as ms timestamp
        expected_large = datetime.fromtimestamp(large_value / 1000, tz=timezone.utc).strftime('%Y-%m-%d %H:%M:%S')
        self.assertEqual(expected_large, fmt(large_value))

        # Test with None timestamp and custom empty_value
        self.assertEqual("Custom Empty", fmt(None, empty_value="Custom Empty"))
        self.assertEqual("", fmt(None, empty_value=""))
        self.assertEqual("N/A", fmt(None, empty_value="N/A"))

        # Test backward compatibility - default should still be 'Never'
        self.assertEqual("Never", fmt(None))

        # Test with datetime object (should not use empty_value)
        dt_obj = datetime(2021, 1, 1, 0, 0, 0, tzinfo=timezone.utc)
        expected = "2021-01-01 00:00:00"
        self.assertEqual(expected, fmt(dt_obj))
        self.assertEqual(expected, fmt(dt_obj, empty_value="Custom Empty"))

    def test_is_timestamp_ms(self):
        # Get current timestamp in seconds and milliseconds
        current_ts = int(datetime.now().timestamp())
        current_ms = current_ts * 1000

        # Test with current timestamps
        self.assertTrue(is_timestamp_ms(current_ms))  # Should be true for ms timestamp
        self.assertFalse(is_timestamp_ms(current_ts))  # Should be false for seconds timestamp

        # Test with future timestamps
        future_sec = current_ts + 1000
        future_ms = future_sec * 1000
        self.assertTrue(is_timestamp_ms(future_ms))
        self.assertFalse(is_timestamp_ms(future_sec))

        # Test with negative values (should be False)
        self.assertFalse(is_timestamp_ms(-1))
        self.assertFalse(is_timestamp_ms(-current_ts))

        # Test with zero value (should be False)
        self.assertFalse(is_timestamp_ms(0))
        self.assertFalse(is_timestamp_ms("0"))

        # Test with very large value that has enough digits to be considered ms
        large_value = 9999999999999  # This should be treated as ms timestamp
        self.assertTrue(is_timestamp_ms(large_value))

        # Test with string representations
        self.assertTrue(is_timestamp_ms(str(current_ms)))  # String ms timestamp
        self.assertFalse(is_timestamp_ms(str(current_ts)))  # String seconds timestamp

        # Test with invalid strings
        self.assertFalse(is_timestamp_ms("not-a-timestamp"))
        self.assertFalse(is_timestamp_ms("123abc"))

    def test_s_fmt(self):
        """Combined test for s_fmt function covering all scenarios"""
        # Test basic functionality from docstring example
        result = s_fmt('Ziegh1xoIlievu0kid7Ub3eiPheegh6Jtichei3AAQuaisa3oob4aeNeAhmai0ch', 3, 3)
        self.assertEqual(result, 'Zie***0ch')

        # Test with a string that's too short to show both beginning and ending
        # When the total length is less than or equal to (show_beginning + show_ending)
        result = s_fmt('12345', 3, 3)  # Length = 5, show_beginning + show_ending = 6
        self.assertEqual(result, '********')  # Should return len(s) + 3 asterisks = 5 + 3 = 8

        # Test with a very short string
        result = s_fmt('12', 1, 1)  # Length = 2, show_beginning + show_ending = 2
        self.assertEqual(result, '*****')  # Should return len(s) + 3 asterisks = 2 + 3 = 5

        # Test with an empty string
        result = s_fmt('', 3, 3)
        self.assertEqual(result, '***')  # Should return len(s) + 3 asterisks = 0 + 3 = 3

        # Test with custom show_beginning parameter
        result = s_fmt('abcdefghijklmnop', 5, 2)
        self.assertEqual(result, 'abcde***op')

        # Test with custom show_ending parameter
        result = s_fmt('abcdefghijklmnopq', 2, 5)
        self.assertEqual(result, 'ab***mnopq')

        # Test with custom show_beginning and show_ending parameters
        result = s_fmt('thisisaverylongstringexample', 4, 4)
        self.assertEqual(result, 'this***mple')

        # Test with minimum values for show_beginning and show_ending
        result = s_fmt('1234567890', 1, 1)
        self.assertEqual(result, '1***0')

        # Test with single character string
        result = s_fmt('a', 1, 1)
        self.assertEqual(result, '****')  # Should return len(s) + 3 asterisks = 1 + 3 = 4

        # Test when show_beginning and show_ending are zero
        result = s_fmt('verylongstring', 0, 0)
        self.assertEqual(result, '***')

        # Test with a long string using default parameters
        long_string = 'A' * 20
        result = s_fmt(long_string)
        self.assertEqual(result, 'AAA***AAA')

    def test_format_number(self):
        """Unified test for format_number function covering all cases."""
        # Test formatting with Decimal values
        self.assertEqual(format_number(Decimal('123.456')), '123.456')
        self.assertEqual(format_number(Decimal('123.450')), '123.45')
        self.assertEqual(format_number(Decimal('123.000')), '123')
        self.assertEqual(format_number(Decimal('0.000')), '0')
        self.assertEqual(format_number(Decimal('123.456789012345')), '123.456789012345')
        self.assertEqual(format_number(Decimal('-123.456')), '-123.456')
        self.assertEqual(format_number(Decimal('-123.450')), '-123.45')
        self.assertEqual(format_number(Decimal('-123.000')), '-123')

        # Test formatting with float values (using string representation to avoid precision issues)
        self.assertEqual(format_number(123.456), '123.456')
        self.assertEqual(format_number(123.450), '123.45')
        self.assertEqual(format_number(123.000), '123')
        self.assertEqual(format_number(25.1), '25.1')
        self.assertEqual(format_number(0.000), '0')
        self.assertEqual(format_number(-25.1), '-25.1')

        # Test formatting with integer values
        self.assertEqual(format_number(123), '123')
        self.assertEqual(format_number(0), '0')
        self.assertEqual(format_number(-12), '-12')

        # Test formatting with None value
        self.assertIsNone(format_number(None))

        # Test formatting with custom precision
        self.assertEqual(format_number(Decimal('123.456789'), precision=2), '123.46')
        self.assertEqual(format_number(Decimal('123.456789'), precision=4), '123.4568')
        self.assertEqual(format_number(Decimal('123.456789'), precision=6), '123.456789')
        self.assertEqual(format_number(Decimal('123.456789'), precision=0), '123')

        # Test edge cases
        # Test very small numbers
        self.assertEqual(format_number(Decimal('0.0000001')), '0.0000001')
        self.assertEqual(format_number(Decimal('0.0000000')), '0')

        # Test very large numbers
        self.assertEqual(format_number(Decimal('123456789.123456789')), '123456789.123456789')


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Utility functions for testing."""
from decimal import Decimal


def check_primitive_types(data):
    """Recursively check that data contains only primitive types and valid nested structures.

    Args:
        data: The data to check, which can be a dict, list, or primitive type.

    Raises:
        AssertionError: If data contains non-primitive types or invalid nested structures.
    """
    if isinstance(data, dict):
        for key, value in data.items():
            assert isinstance(key, str), f"Dict key must be a string, got {type(key)}"
            check_primitive_types(value)
    elif isinstance(data, list):
        for item in data:
            check_primitive_types(item)
    else:
        assert (
            isinstance(data, (str, int, float, bool, Decimal)) or data is None
        ), (f"Invalid type {type(data)}. "
            f"Must be a primitive type (str, int, float, bool, Decimal, None) or a nested structure.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Contains AssetSet class, representing set of assets, and AssetSetError exception class"""
from decimal import Decimal
from typing import List, MappingView, Union


class AssetSetError(Exception):
    """Wrong asset set"""


class AssetSet:
    """Class representing pure set of assets. Allow to compare, subtract, etc"""
    def __init__(self, tickers: Union[List[str], MappingView],
                 amounts: Union[List[Union[float, int, Decimal]], MappingView]):
        # Early check of data
        if not all([isinstance(_, str) for _ in tickers]):
            raise AssetSetError(f'Tickers are wrong, all tickers must be strings! {tickers}')
        elif not all([isinstance(_, (int, float, Decimal)) for _ in amounts]):
            raise AssetSetError(f'Amounts are wrong, all amounts must be numbers! {amounts}')

        # Create main properties
        self.current_index = 0
        self.assets = {}
        self.tickers = [str(_) for _ in tickers]
        self.amounts = [float(_) for _ in amounts]

        # Continue checking
        if len(self.tickers) != len(set(self.tickers)):
            raise AssetSetError(f'Tickers must be unique! {self.tickers}')
        elif len(self.tickers) != len(self.amounts):
            raise AssetSetError(f'Length of tickers not equal to length of assets amounts! '
                                f'{len(self.tickers)}!={len(self.amounts)}')

        # Everything is ok, sort the portfolio by tickers alphabetically if not empty
        if self.tickers:
            self.__sort()

    def __len__(self): return len(self.assets)
    def __contains__(self, item): return item in self.assets

    def __setitem__(self, key, value):
        if key in self.assets:
            idx = self.tickers.index(key)
            self.amounts[idx] = value
            self.assets[key] = value
        else:
            need_sort = (key < self.tickers[-1]) if len(self) else False
            self.tickers.append(key)
            self.amounts.append(value)
            self.assets[key] = value
            if need_sort:
                self.__sort()

    def __delitem__(self, key):
        idx = self.tickers.index(key)
        del self.tickers[idx]
        del self.amounts[idx]
        del self.assets[key]

    def __getitem__(self, item: Union[int, str]):
        if isinstance(item, int):
            return self.amounts[item]
        elif isinstance(item, str):
            return self.assets[item]
        else:
            raise TypeError(f'Wrong key type! Must be int or str, {type(item)} given')

    def __iter__(self):
        self.current_index = 0
        return self

    def __next__(self):
        if self.current_index < len(self):
            v = self.tickers[self.current_index]
            self.current_index += 1
            return v
        raise StopIteration

    def __ne__(self, other): return self.assets != other.assets
    def __eq__(self, other): return self.assets == other.assets

    def __str__(self):
        return str(self.assets)

    def __sub__(self, other):
        tickers = sorted(set(self.tickers + other.tickers))  # combined tickers list, sorted and unique
        left = {_: self.assets[_] if _ in self.assets else 0 for _ in tickers}
        right = {_: other.assets[_] if _ in other.assets else 0 for _ in tickers}
        diff = {_: left[_] - right[_] for _ in tickers}
        return self.__class__.from_dict(diff)

    @property
    def n(self): return len(self)

    def items(self): return self.assets.items()

    def __sort(self):
        self.tickers, self.amounts = zip(*sorted(zip(self.tickers, self.amounts, strict=True)), strict=True)
        self.tickers, self.amounts = list(self.tickers), list(self.amounts)
        self.assets = dict(zip(self.tickers, self.amounts, strict=True))

    @classmethod
    def from_dict(cls, d: dict):
        return cls(d.keys(), d.values())

    def values(self, prices: List[Union[float, int, Decimal]]) -> List[float]:
        return [amount * float(price) for amount, price in zip(self.amounts, prices, strict=True)] \
            if self.amounts and prices else []

    def value(self, prices: List[Union[float, int, Decimal]]) -> float:
        return sum(self.values(prices)) if self.amounts and prices else 0

    def weights(self, prices: List[Union[float, int, Decimal]]) -> List[float]:
        values = self.values(prices)
        total_value = sum(values)
        return [value/total_value for value in values]


if __name__ == "__main__":
    # Create AssetSet of {'BTC': 0.01, 'USDT': 100}
    asset_set = AssetSet(['BTC', 'USDT'], [0.01, 100])
    print(asset_set)

    # Add 0.1 Ethereum to asset set
    asset_set['ETH'] = 0.1
    print(asset_set)

    # Get amount of BTC in asset set (get BTC balance in other words)
    print(asset_set['BTC'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import unittest
from datetime import datetime, timezone
from decimal import Decimal
from typing import Optional, Union
from unittest.mock import MagicMock

from amarket.base import MarketBase, TradingAccountBase
from amarket.constants import ORDER_TYPE_LIMIT, ORDER_TYPE_MARKET, SIDE_BUY, SIDE_SELL
from amarket.interfaces import IOrder
from amarket.local import LocalMarket
from amarket.types import TaggedBalancesDict


class TradingAccount1(TradingAccountBase):
    def get_api_key_permissions(self, update: bool = False) -> dict: pass
    def get_balances(self, update: bool = False) -> TaggedBalancesDict: pass
    def balance(self, ticker: str, tag: str = '') -> Decimal: pass
    def get_open_orders(self, symbol: Optional[str] = None, update: bool = False) -> dict[int, IOrder]: pass
    def is_order_open(self, order: dict | IOrder) -> bool: pass
    def get_orders(self, symbol: str, start_time: Optional[Union[int, float, str]] = None,
                   end_time: Optional[Union[int, float, str]] = None, update: bool = False) -> dict[int, IOrder]: pass

    def normalize_order(self, raw_order_data: dict) -> IOrder: pass

    def get_order(self, symbol: str, order_id: Optional[int] = None, client_order_id: Optional[str] = None,
                  update: bool = False) -> Optional[IOrder]: pass

    def send_order(self, symbol: str, side: SIDE_BUY | SIDE_SELL,
                   order_type: Union[ORDER_TYPE_LIMIT, ORDER_TYPE_MARKET, str], **kwargs) -> IOrder:
        super().send_order(symbol, side, order_type, **kwargs)
        order = MagicMock(spec=IOrder)
        order.symbol = symbol
        order.side = side
        order.order_type = order_type
        order.quantity = Decimal(kwargs.get('quantity', 1))
        order.price = Decimal(kwargs.get('price', 10000))  # hardcode this for market orders testing purposes
        order.order_id = 123456789
        return order


class TradingAccount2(TradingAccountBase):
    def get_api_key_permissions(self, update: bool = False) -> dict: pass
    def get_balances(self, update: bool = False) -> TaggedBalancesDict: pass
    def balance(self, ticker: str, tag: str = '') -> Decimal: pass
    def get_open_orders(self, symbol: Optional[str] = None, update: bool = False) -> dict[int, IOrder]: pass
    def is_order_open(self, order: dict | IOrder) -> bool: pass
    def get_orders(self, symbol: str, start_time: Optional[Union[int, float, str]] = None,
                   end_time: Optional[Union[int, float, str]] = None, update: bool = False) -> dict[int, IOrder]: pass

    def normalize_order(self, raw_order_data: dict) -> IOrder: pass

    def get_order(self, symbol: str, order_id: Optional[int] = None, client_order_id: Optional[str] = None,
                  update: bool = False) -> Optional[IOrder]: pass

    def send_order(self, symbol: str, side: SIDE_BUY | SIDE_SELL,
                   order_type: Union[ORDER_TYPE_LIMIT, ORDER_TYPE_MARKET, str], **kwargs) -> IOrder: pass


class TestTradingAccountBase(unittest.TestCase):

    def setUp(self):
        # Create credentials
        self.api_key = "test_api_key"
        self.api_secret = "test_api_secret"

        # Reset instances on every test start
        with TradingAccountBase._singleton_lock:
            TradingAccountBase._singleton_instances = {}

        # Create new Trading Account on every test start
        self.trading_account = TradingAccount1(self.api_key, self.api_secret)

    def test_singleton(self):
        # Test singleton same credentials
        instance1, instance2 = self.trading_account, TradingAccount1(self.api_key, self.api_secret)
        self.assertIs(instance1, instance2)
        self.assertEqual(1, instance1.n_instances)
        self.assertEqual(1, instance2.n_instances)
        self.assertEqual(1, len(TradingAccountBase._singleton_instances))
        self.assertEqual(1, len(TradingAccount1._singleton_instances))
        self.assertIs(TradingAccountBase._singleton_instances, TradingAccount1._singleton_instances)

        # Test singleton different symbol and timeframe
        instance3 = TradingAccount1("test_api_key_another", "test_api_secret_another")
        self.assertIs(instance1, instance2)
        self.assertIsNot(instance1, instance3)
        self.assertIsNot(instance2, instance3)
        self.assertEqual(2, len(TradingAccountBase._singleton_instances))
        self.assertEqual(2, len(TradingAccount1._singleton_instances))
        self.assertEqual(2, instance1.n_instances)
        self.assertEqual(2, instance2.n_instances)
        self.assertEqual(2, instance3.n_instances)
        self.assertIs(TradingAccountBase._singleton_instances, TradingAccount1._singleton_instances)
        self.assertIs(TradingAccountBase._singleton_instances, TradingAccount2._singleton_instances)

        # Test tbe singleton logic works right in case of another child
        instance3 = TradingAccount2(self.api_key, self.api_secret)
        self.assertIs(instance1, instance2)
        self.assertIsNot(instance3, instance1)
        self.assertIsNot(instance3, instance2)
        self.assertEqual(3, len(TradingAccountBase._singleton_instances))
        self.assertEqual(3, len(TradingAccount1._singleton_instances))
        self.assertEqual(3, len(TradingAccount2._singleton_instances))
        self.assertEqual(2, instance1.n_instances)
        self.assertEqual(2, instance2.n_instances)
        self.assertEqual(1, instance3.n_instances)
        self.assertIs(TradingAccountBase._singleton_instances, TradingAccount1._singleton_instances)
        self.assertIs(TradingAccountBase._singleton_instances, TradingAccount2._singleton_instances)

    def test_send_order(self):
        # Test valid LIMIT order
        with self.assertRaises(ValueError) as context:
            self.trading_account.send_order("BTCUSDT", SIDE_BUY, ORDER_TYPE_LIMIT)
        self.assertEqual(str(context.exception), "A LIMIT order requires both quantity and price.")

        # Test valid MARKET order with quantity
        with self.assertRaises(ValueError) as context:
            self.trading_account.send_order("BTCUSDT", SIDE_BUY, ORDER_TYPE_MARKET)
        self.assertEqual(str(context.exception), "A MARKET order requires either quantity or quote_quantity.")

        # Testing different orders
        for args, kwargs in [
            (("BTCUSDT", SIDE_SELL, ORDER_TYPE_LIMIT), {'quantity': 1, 'price': 100}),  # valid LIMIT order
            (("BTCUSDT", SIDE_BUY, ORDER_TYPE_MARKET), {'quantity': 1}),  # valid MARKET order with quantity
            (("ETHUSDT", SIDE_BUY, ORDER_TYPE_MARKET), {'quantity': 1}),  # valid MARKET order with quote_quantity
        ]:
            symbol, side, order_type = args
            order = self.trading_account.send_order(symbol, side, order_type, **kwargs)
            self.assertIsInstance(order, IOrder)
            self.assertEqual(order.order_id, 123456789)
            self.assertEqual(order.symbol, symbol)
            self.assertEqual(order.side, side)
            self.assertEqual(order.order_type, order_type)
            self.assertEqual(order.price, Decimal(kwargs.get('price', 10000)))  # Default price for market order
            self.assertIsInstance(order.quantity, Decimal)

    def test_test_order(self):
        # Test valid parameters
        self.assertTrue(self.trading_account.test_order("BTCUSDT", SIDE_BUY, "MARKET"))
        self.assertTrue(self.trading_account.test_order("BTCUSDT", SIDE_SELL, "LIMIT", price="1000"))

        # Test invalid symbol
        with self.assertRaises(ValueError):
            self.trading_account.test_order("", SIDE_BUY, "MARKET")
        with self.assertRaises(ValueError):
            self.trading_account.test_order(None, SIDE_BUY, "MARKET")

        # Test invalid side
        with self.assertRaises(ValueError):
            self.trading_account.test_order("BTCUSDT", 2, "MARKET")

        # Test invalid order type
        with self.assertRaises(ValueError):
            self.trading_account.test_order("BTCUSDT", SIDE_BUY, "STOP")

    def test_api_usage_statistics(self):
        """Test API usage statistics functionality"""
        # Test _update_api_usage with various headers
        test_headers = {
            'x-mbx-used-weight': '10',
            'x-mbx-used-weight-1m': '555',
            'retry-after': '120',
            'date': 'Mon, 01 Jan 2023 00:00:00 GMT'
        }
        self.trading_account._update_api_usage(test_headers, status_code=200)

        # Check that api_usage was updated
        api_usage = self.trading_account.api_usage
        self.assertEqual(api_usage['used_weight'], 10)
        self.assertEqual(api_usage['used_weight_1m'], 555)
        self.assertEqual(api_usage['retry_after'], 120)
        self.assertTrue(api_usage['timestamp'] > 0)
        self.assertEqual(api_usage['status_code'], 200)

        # Check that statistics were updated
        # The exact index depends on the current time, but we can check that
        # some value was updated in the statistics
        non_zero_count = sum(1 for v in self.trading_account._api_usage_statistics if v > 0)
        self.assertGreater(non_zero_count, 0)
        self.assertEqual(non_zero_count, 1)

        # Test multiple updates to ensure statistics rotation
        trading_account = TradingAccount2(self.api_key, self.api_secret, api_usage_storage_size=5)
        for i in range(10):
            expected = i * 2 + 1  # avoid zero value because later we count non-zero values
            test_headers = {
                'x-mbx-used-weight': str(i * 5),
                'x-mbx-used-weight-1m': str(expected),
                'date': f'Mon, 01 Jan 2023 00:{str(i).zfill(2)}:00 GMT'
            }
            self.trading_account._update_api_usage(test_headers, status_code=200)
            self.assertEqual(expected, next(reversed(self.trading_account._api_usage_statistics.values())))
            self.assertEqual(self.trading_account.api_usage['status_code'], 200)

            trading_account._update_api_usage(test_headers, status_code=200)
            self.assertEqual(expected, next(reversed(trading_account._api_usage_statistics.values())))
            self.assertEqual(trading_account.api_usage['status_code'], 200)

        # Check that statistics are properly rotated and updated
        non_zero_count = sum(1 for v in self.trading_account._api_usage_statistics if v > 0)
        self.assertGreater(non_zero_count, 0)
        self.assertEqual(non_zero_count, 10)

        # Check that only the last 5 entries are kept
        self.assertEqual(len(trading_account._api_usage_statistics), 5)
        self.assertEqual(len(self.trading_account._api_usage_statistics), 10)  # this must kept untouched

        # Check that the oldest entries were rotated out
        # The timestamps should be from minutes 5-9 (not 0-4)
        base_timestamp = 1672531200  # Mon, 01 Jan 2023 00:00:00 GMT
        expected_timestamps = [base_timestamp + (i * 60) for i in range(5, 10)]  # 5, 6, 7, 8, 9 minutes
        self.assertListEqual(list(trading_account._api_usage_statistics.keys()), expected_timestamps)

        # Check that values correspond to the last 5 updates (5-9)
        values = list(trading_account._api_usage_statistics.values())
        expected_values = [i * 2 + 1 for i in range(5, 10)]  # 11, 13, 15, 17, 19
        self.assertEqual(values, expected_values)

    def test_get_api_usage_statistics(self):
        """Test the get_api_usage_statistics method"""
        # Fill statistics
        for i in range(10):
            test_headers = {
                'x-mbx-used-weight': str(i * 5),
                'x-mbx-used-weight-1m': str(i * 2 + 1),
                'date': f'Mon, 01 Jan 2023 00:{str(i).zfill(2)}:00 GMT'
            }
            self.trading_account._update_api_usage(test_headers, status_code=200)

        # Test with default retention period (7200 minutes)
        stats = self.trading_account.get_api_usage_statistics(fmt=None)
        self.assertEqual(len(stats), 10)
        self.assertTrue(all(isinstance(k, int) for k in stats.keys()))  # All keys should be timestamps
        self.assertTrue(all(isinstance(v, int) for v in stats.values()))  # All values should be integers

        # Test with formatted dates
        formatted_stats = self.trading_account.get_api_usage_statistics()
        self.assertEqual(len(formatted_stats), 10)
        self.assertTrue(all(isinstance(k, str) for k in formatted_stats.keys()))  # All keys should be strings
        self.assertTrue(all(isinstance(v, int) for v in formatted_stats.values()))  # All values should be integers

        # Test that keys are properly formatted when
        for key in formatted_stats.keys():
            self.assertRegex(key, r'\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}')  # YYYY-MM-DD HH:MM:SS format

        # Test that statistics are consistent between raw and formatted versions
        raw_stats = self.trading_account.get_api_usage_statistics(fmt=None)
        formatted_stats = self.trading_account.get_api_usage_statistics()

        # Convert formatted keys back to timestamps for comparison
        timestamp_keys = []
        for formatted_key, value in formatted_stats.items():
            timestamp = datetime.strptime(formatted_key, "%Y-%m-%d %H:%M:%S").replace(tzinfo=timezone.utc).timestamp()
            timestamp_keys.append(timestamp)

        # Sort both sets of keys for comparison
        sorted_raw_keys = sorted(raw_stats.keys())
        sorted_timestamp_keys = sorted(timestamp_keys)

        # Check that all values are the same
        self.assertEqual(len(raw_stats), len(formatted_stats))
        self.assertEqual(sorted_raw_keys, sorted_timestamp_keys)
        for raw_key, formatted_key in zip(sorted_raw_keys, sorted_timestamp_keys):
            self.assertEqual(
                raw_stats[raw_key],
                formatted_stats[datetime.fromtimestamp(formatted_key, tz=timezone.utc).strftime("%Y-%m-%d %H:%M:%S")]
            )

        # Test with some actual API usage data
        test_headers = {
            'x-mbx-used-weight': '10',
            'x-mbx-used-weight-1m': '555',
            'retry-after': '120',
            'date': 'Mon, 01 Jan 2023 00:00:00 GMT'
        }
        self.trading_account._update_api_usage(test_headers, status_code=200)

        # Get statistics after update
        stats_after_update = self.trading_account.get_api_usage_statistics(fmt=None)
        formatted_stats_after_update = self.trading_account.get_api_usage_statistics()

        # Check that some values are non-zero
        non_zero_count = sum(1 for v in stats_after_update.values() if v > 0)
        self.assertGreater(non_zero_count, 0)

        # Check that the same non-zero count appears in both versions
        formatted_non_zero_count = sum(1 for v in formatted_stats_after_update.values() if v > 0)
        self.assertEqual(non_zero_count, formatted_non_zero_count)

    def test_new_simple_trade(self):
        """Test new_simple_trade method with various scenarios"""
        # Create respective market for order roundings
        market = LocalMarket(exchange="binance", market_type="spot", symbol="BTCUSDT", base_ticker="BTC",
                             quote_ticker="USDT", lot_precision=4, price_precision=2)

        # Manually register the market in MarketBase._singleton_instances as LocalMarket not participate in singleton
        k = (type(market), market.symbol)
        MarketBase._singleton_instances[k] = market

        # Create account and set exchange and market_type
        self.trading_account._exchange = "binance"
        self.trading_account._market_type = "spot"

        # Test invalid symbol
        with self.assertRaises(ValueError) as context:
            self.trading_account.new_simple_trade("", SIDE_BUY, Decimal('0.1'), Decimal('0.05'),
                                                  exit_pct=Decimal('0.02'))
        self.assertEqual(str(context.exception), "Symbol must be a non-empty string.")

        # Test invalid side
        with self.assertRaises(ValueError) as context:
            self.trading_account.new_simple_trade("BTCUSDT", 2, Decimal('0.1'), 0.05, exit_pct=Decimal('0.02'))
        self.assertEqual(str(context.exception), f"Side must be {SIDE_BUY} for BUY or {SIDE_SELL} for SELL.")

        # Test invalid lot_entry (non-positive)
        with self.assertRaises(ValueError) as context:
            self.trading_account.new_simple_trade("BTCUSDT", SIDE_BUY, Decimal('0'), Decimal('0.05'),
                                                  exit_pct=Decimal('0.02'))
        self.assertEqual(str(context.exception), "lot_entry must be positive.")

        # Test invalid lot_exit (non-positive)
        with self.assertRaises(ValueError) as context:
            self.trading_account.new_simple_trade("BTCUSDT", SIDE_BUY, Decimal('0.1'), Decimal('0'),
                                                  exit_pct=Decimal('0.02'))
        self.assertEqual(str(context.exception), "lot_exit must be positive.")

        # Test invalid exit_pct (non-positive)
        with self.assertRaises(ValueError) as context:
            self.trading_account.new_simple_trade("BTCUSDT", SIDE_BUY, Decimal('0.1'), Decimal('0.05'),
                                                  exit_pct=Decimal('0'))
        self.assertEqual(str(context.exception), "exit_pct must be positive.")

        # Test successful trade execution with exit_pct
        orders = self.trading_account.new_simple_trade("BTCUSDT", SIDE_BUY, '0.1', Decimal('0.05'),
                                                       exit_pct=Decimal('0.02'))
        self.assertEqual(len(orders), 2)

        # Test validation: neither exit_pct nor exit_price provided
        with self.assertRaises(ValueError) as context:
            self.trading_account.new_simple_trade("BTCUSDT", SIDE_BUY, Decimal('0.1'), Decimal('0.05'))
        self.assertEqual(str(context.exception), "Either exit_pct or exit_price must be provided.")

        # Test validation: both exit_pct and exit_price provided
        with self.assertRaises(ValueError) as context:
            self.trading_account.new_simple_trade("BTCUSDT", SIDE_BUY, Decimal('0.1'), Decimal('0.05'),
                                                  exit_pct=Decimal('0.02'), exit_price=Decimal('10200'))
        self.assertEqual(str(context.exception), "Only one of exit_pct or exit_price can be provided.")

        # Test successful trade execution with exit_pct
        orders = self.trading_account.new_simple_trade("BTCUSDT", SIDE_BUY, Decimal('0.1'), Decimal('0.05'),
                                                       exit_pct=Decimal('0.02'))
        self.assertEqual(len(orders), 2)
        entry_order, exit_order = orders

        # Verify both orders have the expected properties
        # Check entry order (MARKET order)
        self.assertIsInstance(entry_order, MagicMock)
        self.assertEqual(entry_order.order_type, ORDER_TYPE_MARKET)
        self.assertEqual(entry_order.side, SIDE_BUY)
        self.assertEqual(entry_order.quantity, Decimal('0.1'))
        self.assertEqual(entry_order.price, Decimal('10000'))  # Default market price

        # Check exit take order (LIMIT order) with calculated price
        self.assertIsInstance(exit_order, MagicMock)
        self.assertEqual(exit_order.order_type, ORDER_TYPE_LIMIT)
        self.assertEqual(exit_order.side, SIDE_SELL)  # Opposite side
        self.assertEqual(exit_order.quantity, Decimal('0.05'))
        expected_price = Decimal('10000') + Decimal('10000') * Decimal('0.02')
        self.assertEqual(exit_order.price, expected_price)

        # Test successful trade execution with exit_price
        orders = self.trading_account.new_simple_trade("BTCUSDT", SIDE_BUY, Decimal('0.1'), Decimal('0.05'),
                                                       exit_price=Decimal('10200'))
        self.assertEqual(len(orders), 2)
        entry_order, exit_order = orders

        # Verify both orders have the expected properties
        # Check entry order (MARKET order)
        self.assertIsInstance(entry_order, MagicMock)
        self.assertEqual(entry_order.order_type, ORDER_TYPE_MARKET)
        self.assertEqual(entry_order.side, SIDE_BUY)
        self.assertEqual(entry_order.quantity, Decimal('0.1'))
        self.assertEqual(entry_order.price, Decimal('10000'))  # Default market price

        # Check exit take order (LIMIT order) with fixed price
        self.assertIsInstance(exit_order, MagicMock)
        self.assertEqual(exit_order.order_type, ORDER_TYPE_LIMIT)
        self.assertEqual(exit_order.side, SIDE_SELL)  # Opposite side
        self.assertEqual(exit_order.quantity, Decimal('0.05'))
        self.assertEqual(exit_order.price, Decimal('10200'))  # Fixed price from exit_price

        # Test exit-take price calculation for LONG trades (SIDE_BUY)
        orders = self.trading_account.new_simple_trade(
            "BTCUSDT", SIDE_BUY, Decimal('0.1'), Decimal('0.05'),
            exit_pct=Decimal('0.02'))  # 2% take profit
        _, exit_order = orders

        # For LONG trades, exit-take price should be entry_price + (entry_price * pct)
        expected_price = Decimal('10000') + Decimal('10000') * Decimal('0.02')
        self.assertEqual(exit_order.price, expected_price)

        # Test exit-take price calculation for SHORT trades (SIDE_SELL)
        orders = self.trading_account.new_simple_trade(
            "BTCUSDT", SIDE_SELL, Decimal('0.1'), Decimal('0.05'),
            exit_pct=Decimal('0.02'))  # 2% take profit
        _, exit_order = orders

        # For SHORT trades, exit-take price should be entry_price - (entry_price * pct)
        expected_price = Decimal('10000') - Decimal('10000') * Decimal('0.02')
        self.assertEqual(exit_order.price, expected_price)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Contains classes representing position"""
from datetime import datetime, timezone

__package__ = 'amarket'

from .symbol_account_mixin import SymbolAccountMixin
from .symbol_mixin import SymbolMixin


class PositionError(Exception):
    """Wrong position"""
    pass


def validate(func):
    def func_wrapper(*args, **kwargs):  # args must be: self, quantity, price, etc.; kwargs may contain currency, limit
        self = args[0]
        quantity = args[1]
        price = args[2]
        cname = self.__class__.__name__
        fname = func.__name__

        if self.is_closed:
            raise PositionError("{}.{}({}, {}): Position is closed".format(cname, fname, quantity, price))

        if not (isinstance(quantity, (int, float)) and isinstance(price, (int, float))):
            raise PositionError("{}.{}({}, {}): Both quantity and price must be numeric"
                                .format(cname, fname, quantity, price))

        if quantity <= 0 or price <= 0:
            raise PositionError("{}.{}({}, {}): Both quantity and price have to be greater than zero"
                                .format(cname, fname, quantity, price))

        if 'limit' in kwargs:
            if not isinstance(kwargs['limit'], (int, float)) or kwargs['limit'] < 0:
                raise PositionError(f"{cname}.{fname}({quantity}, {price}, limit={kwargs['limit']}): "
                                    f"Limit price must be the number and greater than zero")

        if 'currency' in kwargs:
            currency = kwargs['currency']
            if currency not in (Position.CURRENCY_BASE, Position.CURRENCY_QUOTE):
                raise PositionError("{}.{}({}, {}, currency={}): Currency must be '{}' or '{}'"
                                    .format(cname, fname, quantity, price, currency,
                                            Position.CURRENCY_BASE, Position.CURRENCY_QUOTE))
        else:
            currency = Position.CURRENCY_BASE

        if fname == 'decrease':
            qty = quantity if currency == Position.CURRENCY_BASE else quantity / price
            if qty > self.direction * self.base_quantity:
                raise PositionError('{}.{}({}, {}, currency={}): Can not decrease position because it has no such '
                                    'currency quantity (there are only {} of base, {} of quote). '
                                    'Try Position.close({})'.format(cname, fname, quantity, price, currency,
                                                                    self.base_quantity, self.quote_quantity, price))

        return func(*args, **kwargs)
    return func_wrapper


class Position(SymbolAccountMixin, SymbolMixin):
    """Class representing market position. Support only market orders (may say it processes only trades)"""
    DIRECTION_LONG = 1
    DIRECTION_SHORT = -1

    STATUS_NEW = 0  # there are no trades executed, pending opening
    STATUS_OPENED = 1  # there are executed trades
    STATUS_CLOSED = 7  # position closed, no further position adjustment allowed

    @property
    def direction(self): return self.__direction

    @property
    def average_price(self): return -1 * self.quote_quantity / self.base_quantity if self.base_quantity != 0 else 0

    @property
    def status(self): return self._status

    @property
    def created(self): return self._created

    @property
    def updated(self): return self._updated

    @property
    def opened(self): return self._opened

    @property
    def closed(self): return self._closed

    @property
    def has_trades(self) -> bool:
        """@:returns True if position has any trades, False otherwise"""
        return True if (self.base_quantity or self.quote_quantity) and not self.is_new else False

    @property
    def is_new(self) -> bool:
        """@:returns True if position is new"""
        return self.status == self.STATUS_NEW

    @property
    def is_closed(self) -> bool:
        """@:returns True if position is closed"""
        return self.status == self.STATUS_CLOSED

    def __init__(self, direction, *args, created=None, **kwargs):
        # Check inputs
        if direction != self.DIRECTION_LONG and direction != self.DIRECTION_SHORT:
            raise PositionError('Position direction must be Position.BUY or Position.SELL')
        self.__direction = direction
        self._status = self.STATUS_NEW
        self._created = datetime.now(timezone.utc) if created is None or not isinstance(created, datetime) else created
        self._updated = self._created
        self._opened = None
        self._closed = None
        super().__init__(*args, **kwargs)

    def to_dict(self):
        return {
            'direction': self.direction,
            'base_quantity': self.base_quantity,
            'quote_quantity': self.quote_quantity,
            'status': self.status,
            'created': self.created,
            'updated': self.updated,
            'opened': self.opened,
            'closed': self.closed,
        }

    @classmethod
    def from_dict(cls, d: dict):
        for k in ['direction', 'base_quantity', 'quote_quantity', 'status']:
            if k not in d:
                raise PositionError(f'{cls.__name__}.from_dict(...): Dictionary must contain keys: '
                                    "'direction', 'base_quantity', 'quote_quantity' and 'status'")

        entity = cls(d['direction'], created=d['created'] if 'created' in d else None)
        entity._status = d['status']

        if entity.is_new:
            if d['base_quantity'] != 0 or d['quote_quantity'] != 0:
                raise PositionError(f'{cls.__name__}.from_dict(...): '
                                    f'Base and quote quantity must be zero for new position!')
        elif entity.is_closed and d['base_quantity'] > 0:
            raise PositionError(f'{cls.__name__}.from_dict(...): Base quantity must be zero for closed position!')
        elif d['base_quantity'] < 0 and d['quote_quantity'] < 0:
            raise PositionError(f'{cls.__name__}.from_dict(...): Base and quote quantities can not be negative both!')

        entity._base_quantity = d['base_quantity']
        entity._quote_quantity = d['quote_quantity']

        if 'opened' in d:
            if d['opened'] < entity.created:
                raise PositionError(f'{cls.__name__}.from_dict(...): Open time is earlier than create time!')
            if entity.is_new:
                raise PositionError(f'{cls.__name__}.from_dict(...): Open time is given but status is new!')
            entity._opened = d['opened']
            entity._updated = d['opened']

        if 'closed' in d:
            if not entity.is_closed:
                raise PositionError(f'{cls.__name__}.from_dict(...): Close time is given but status is not closed!')
            else:
                if d['closed'] < entity.opened:
                    raise PositionError(f'{cls.__name__}.from_dict(...): Close time is earlier than open time!')
            entity._closed = d['closed']
            entity._updated = d['closed']

        return entity

    @classmethod
    def create_and_open(cls, direction, quantity, price, currency='base', opened=None):
        entity = cls(direction, created=opened)
        entity.increase(quantity, price, currency=currency, updated=opened)
        return entity

    def open(self, quantity, price, currency='base', opened=None, direction=None):
        if not self.base_quantity and not self.quote_quantity and self.is_new:
            if direction == self.DIRECTION_LONG or direction == self.DIRECTION_SHORT:
                self.__direction = direction
            self.increase(quantity, price, currency=currency, updated=opened)
        else:
            raise PositionError(f'{self.__class__.__name__}.open({quantity}, {price}...): '
                                f'Try to open not new (status={self.status}) position!')

    @validate
    def increase(self, quantity, price, currency='base', updated=None):
        if currency == self.CURRENCY_BASE:
            self._base_quantity += self.direction * quantity
            self._quote_quantity -= self.direction * quantity * price
        else:
            self._base_quantity += self.direction * quantity / price
            self._quote_quantity -= self.direction * quantity
        self._status = self.STATUS_OPENED
        updated = datetime.now(timezone.utc) if updated is None or not isinstance(updated, datetime) else updated
        self._opened = updated if self._opened is None else self._opened  # first increase() call, set self._opened
        self._updated = updated

    @validate
    def decrease(self, quantity, price, currency='base', updated=None):
        if currency == self.CURRENCY_BASE:
            self._base_quantity -= self.direction * quantity
            self._quote_quantity += self.direction * quantity * price
        else:
            self._base_quantity -= self.direction * quantity / price
            self._quote_quantity += self.direction * quantity
        self._status = self.STATUS_OPENED
        self._updated = datetime.now(timezone.utc) if updated is None or not isinstance(updated, datetime) else updated

    def close(self, price, closed=None):
        if not isinstance(price, (int, float)) or price <= 0:
            raise PositionError(f'{self.__class__.__name__}.close({price}): '
                                f'Close price must be the number and greater than zero')
        if not self.is_new and self.has_trades and not self.is_closed:
            self.decrease(abs(self._base_quantity), price, currency=self.CURRENCY_BASE, updated=closed)
        self._status = self.STATUS_CLOSED
        self._closed = datetime.now(timezone.utc) if closed is None or not isinstance(closed, datetime) else closed

    def calculate_profit_loss(self, current_price):
        return (current_price - self.average_price) * self.base_quantity

    def calculate_pnl(self, current_price):
        return (current_price / self.average_price - 1 if self.average_price != 0 else 0) * self.direction


class PositionLimit(Position):
    """Position class with limit orders support"""
    # STATUS_NEW: inherited, meaning: there are no trades executed, pending opening (limit orders may exists)
    # STATUS_OPENED: inherited, meaning: there are executed trades and no limit orders waiting fulfillment
    # STATUS_CLOSED: inherited, meaning: position closed, no new orders allowed
    STATUS_PENDING = 2  # there are executed trades and there are limit orders waiting fulfillment

    # Unfilled limit order data
    limit_price: 0
    limit_side: None  # 1 for BUY, -1 for SELL (similar to Position's direction), None - no limit order
    limit_base_quantity: 0

    def __init__(self, direction):
        super().__init__(direction)
        self.cancel_limit()

    def has_limit(self): return False if self.limit_side is None else True  # True if position has unfilled limit order

    def cancel_limit(self):
        self.limit_price = 0
        self.limit_side = None
        self.limit_base_quantity = 0

    def fill_limit(self):
        if self.direction == self.limit_side:
            self.increase(self.limit_base_quantity, self.limit_price, currency=self.CURRENCY_BASE, limit=0)
        else:
            self.decrease(self.limit_base_quantity, self.limit_price, currency=self.CURRENCY_BASE, limit=0)
        self.cancel_limit()

    def to_dict(self):
        d = super().to_dict()
        d['limit_price'] = self.limit_price
        d['limit_side'] = self.limit_side
        d['limit_base_quantity'] = self.limit_base_quantity
        return d

    @validate
    def increase(self, quantity, price, currency='base', limit=0):
        limit_filled = False
        if (self.direction == self.DIRECTION_LONG and limit >= price) or \
                (self.direction == self.DIRECTION_SHORT and limit <= price):
            limit_filled = True

        if not limit or limit_filled:
            return super().increase(quantity, price, currency=currency)
        else:
            self.limit_side = self.direction
            self.limit_price = limit
            self.limit_base_quantity = quantity if currency == self.CURRENCY_BASE else quantity / limit
            self._status = self.STATUS_PENDING if self.has_trades else self.STATUS_NEW

    @validate
    def decrease(self, quantity, price, currency='base', limit=0):
        limit_filled = False
        if (self.direction == self.DIRECTION_LONG and limit <= price) or \
                (self.direction == self.DIRECTION_SHORT and limit >= price):
            limit_filled = True

        if not limit or limit_filled:
            return super().decrease(quantity, price, currency=currency)
        else:
            self.limit_side = -self.direction
            self.limit_price = limit
            self.limit_base_quantity = quantity if currency == self.CURRENCY_BASE else quantity / limit
            self._status = self.STATUS_PENDING if self.has_trades else self.STATUS_NEW


if __name__ == '__main__':
    print('Testing class Position')
    print('======================')
    for direction in (Position.DIRECTION_LONG, Position.DIRECTION_SHORT):
        for currency in (Position.CURRENCY_BASE, Position.CURRENCY_QUOTE):
            qty = 0.001 if currency == 'base' else 20
            print('Direction: {}, Currency: {}, {} units'
                  .format('LONG' if direction == Position.DIRECTION_LONG else 'SHORT', currency.upper(), qty))
            pos = Position(direction)
            print(f'Initiated: {pos.to_dict()}, AvgPrice: {pos.average_price}, has trades: {pos.has_trades}')
            pos.increase(qty, 20000, currency=currency)
            print(f'Increased: {pos.to_dict()}, AvgPrice: {pos.average_price}, has trades: {pos.has_trades}')
            pos.increase(qty, 25000, currency=currency)
            print(f'Increased: {pos.to_dict()}, AvgPrice: {pos.average_price}, has trades: {pos.has_trades}')
            pos.decrease(qty, 30000, currency=currency)
            print(f'Decreased: {pos.to_dict()}, AvgPrice: {pos.average_price}, has trades: {pos.has_trades}')
            pos.close(29000)
            print(f'Closed:    {pos.to_dict()}, AvgPrice: {pos.average_price}, has trades: {pos.has_trades}')
            d = {'direction': 1, 'base_quantity': 0.001, 'quote_quantity': -20.0, 'status': 1}

            pos = Position.from_dict(d)
            print("From dict: {}, AvgPrice: {}, has trades: {}, equal: {}"
                  .format(pos.to_dict(), pos.average_price, pos.has_trades, d == pos.to_dict()))

            pos = Position(direction)
            pos.open(qty, 20000, currency=currency, direction=-direction)
            print(f'Opened:    {pos.to_dict()}, AvgPrice: {pos.average_price}, has trades: {pos.has_trades}')
            pos.open(qty, 20000, currency=currency, direction=direction)
            print(f'Opened-2:  {pos.to_dict()}, AvgPrice: {pos.average_price}, has trades: {pos.has_trades}')

            pos = Position.create_and_open(direction, qty, 20000, currency=currency)
            print(f'Created:   {pos.to_dict()}, AvgPrice: {pos.average_price}, has trades: {pos.has_trades}')

            print()

    print('Testing class PositionLimit')
    print('===========================')
    pos = PositionLimit(PositionLimit.DIRECTION_LONG)
    print('Initiated: {}, AvgPrice: {}, has trades: {}'.format(pos.to_dict(), pos.average_price, pos.has_trades))
    pos.increase(qty, 20000, currency='quote', limit=21000)
    print('Increased: {}, AvgPrice: {}, has trades: {}'.format(pos.to_dict(), pos.average_price, pos.has_trades))
    pos.increase(qty, 21000, currency='quote', limit=19000)
    print('Limit    : {}, AvgPrice: {}, has trades: {}'.format(pos.to_dict(), pos.average_price, pos.has_trades))
    pos.fill_limit()
    print('Filling  : {}, AvgPrice: {}, has trades: {}'.format(pos.to_dict(), pos.average_price, pos.has_trades))



import unittest

from amarket.base import MarketBase
from amarket.constants.market_type import MarketType
from amarket.local import LocalMarket


class TestLocalMarket(unittest.TestCase):
    """Test cases for the LocalMarket class."""

    def setUp(self):
        """Set up test fixtures."""
        self.exchange = 'Exchange_1'
        self.market_type = MarketType.SPOT
        self.symbol = 'BTCUSDT'
        self.base_ticker = 'BTC'
        self.quote_ticker = 'USDT'
        self.lot_precision = 4
        self.price_precision = 2
        self.market = LocalMarket(self.exchange, self.market_type, self.symbol,
                                  base_ticker=self.base_ticker, quote_ticker=self.quote_ticker,
                                  lot_precision=self.lot_precision, price_precision=self.price_precision)

    def test_singleton_disabled(self):
        """Test that singleton behavior is disabled."""
        market2 = LocalMarket(self.exchange, self.market_type, self.symbol,
                              base_ticker=self.base_ticker, quote_ticker=self.quote_ticker,
                              lot_precision=self.lot_precision, price_precision=self.price_precision)
        self.assertIsNot(self.market, market2)
        self.assertDictEqual(self.market._singleton_instances, {})
        self.assertDictEqual(market2._singleton_instances, {})
        self.assertDictEqual(LocalMarket._singleton_instances, {})
        self.assertDictEqual(MarketBase._singleton_instances, {})

    def test_properties(self):
        """Test that properties are set correctly."""
        self.assertEqual(self.market.exchange, self.exchange)
        self.assertEqual(self.market.market_type, self.market_type)
        self.assertEqual(self.market.symbol, self.symbol.upper())
        self.assertEqual(self.market.base_ticker, self.base_ticker)
        self.assertEqual(self.market.quote_ticker, self.quote_ticker)
        self.assertEqual(self.market.lot_precision, self.lot_precision)
        self.assertEqual(self.market.price_precision, self.price_precision)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Contains SymbolMixin and SymbolError classes that implements basic symbol representation"""
from datetime import datetime, timezone
from decimal import Decimal
from math import ceil, floor
from time import time
from typing import Literal, Type, Union


class SymbolError(Exception):
    """Symbol error"""


class SymbolMixin:
    CURRENCY_BASE = 'base'
    CURRENCY_QUOTE = 'quote'

    @property
    def base_ticker(self) -> str: return self.__base_ticker

    @property
    def quote_ticker(self) -> str: return self.__quote_ticker

    @property
    def symbol(self) -> str: return self.__symbol

    @property
    def price(self) -> float: return self.__price

    @price.setter
    def price(self, value: Union[int, float, Decimal]):
        if isinstance(value, (int, float, Decimal)) and value > 0:
            self.__price = float(value)
            self.__price_timestamp = time()
        else:
            raise SymbolError('Price value must be positive non-zero int, float or decimal!')

    @property
    def price_timestamp(self) -> int:
        return round(self.__price_timestamp)

    @price_timestamp.setter
    def price_timestamp(self, value: Union[int, float, Decimal]):
        if isinstance(value, (int, float, Decimal)) and value > 0:
            self.__price_timestamp = float(value)
        else:
            raise SymbolError('Price timestamp value must be positive non-zero int, float or decimal!')

    @property
    def price_timestamp_ms(self) -> int:
        return round(self.__price_timestamp * 1000)

    @property
    def price_datetime(self) -> datetime:
        return datetime.fromtimestamp(self.__price_timestamp, tz=timezone.utc)

    def __init__(self, *args, base_ticker='base', quote_ticker='quote', symbol='', ticker_max_length=10,
                 price=0.0, price_timestamp=0.0,
                 filter_max_quantity=0.0, filter_min_quantity=0.0, filter_step_size=0.0, filter_min_notional=0.0,
                 **kwargs):
        if not base_ticker or not quote_ticker:
            raise SymbolError('Base and quote tickers must be both set!')
        elif len(base_ticker) > ticker_max_length:
            raise SymbolError(f'Base ticker {base_ticker} exceed ticker_max_length={ticker_max_length}!')
        elif len(quote_ticker) > ticker_max_length:
            raise SymbolError(f'Quote ticker {quote_ticker} exceed ticker_max_length={ticker_max_length}!')

        self.__base_ticker = base_ticker.upper()
        self.__quote_ticker = quote_ticker.upper()
        self.__symbol = symbol if symbol and isinstance(symbol, str) else self.base_ticker + self.quote_ticker
        self.__price = price
        self.__price_timestamp = price_timestamp
        self.filter_max_quantity = Decimal(filter_max_quantity)
        self.filter_min_quantity = Decimal(filter_min_quantity)
        self.filter_step_size = Decimal(filter_step_size)
        self.filter_min_notional = Decimal(filter_min_notional)

        if self.__class__.mro() != [self.__class__, object]:  # call parent(s).__init__ if it not 'object' only
            super().__init__(*args, **kwargs)

    def filter_quantity(self, quantity: Union[int, float, Decimal], price: Union[int, float, Decimal] = 0,
                        rounding: Literal["round", "ceil", "floor"] = 'round',
                        min_quantity_pct: float = 0.5, min_notional_pct: float = 0.5,
                        return_as: Union[Type[float], Type[Decimal], Type[int], Type[str]] = float) \
            -> Union[float, Decimal, int, str]:
        """Adjust given quantity to match given filters.
        For some rounding rules explanations, @see https://stackoverflow.com/questions/68199013
        :param quantity: quantity to adjust
        :param rounding: rounding rules (rounding function name to use)
        :param price: price to use calculating MIN_NOTIONAL filter
        :param min_quantity_pct: quantity would be adjusted to match LOT_SIZE.minQty filter, if the resulting
            quantity be great or equal to min_quantity_pct * self.min_quantity, but less than self.min_quantity
        :param min_notional_pct: quantity would be adjusted to match MIN_NOTIONAL filter, if the resulting
            quote quantity be great or equal to min_notional_pct * self.min_notional, but less than self.min_notional
        :param return_as: return as given type
        :return: the adjusted quantity, matching all filters, or zero
        :raise: ValueError if given data not match some criteria (look code)
        """
        if quantity <= 0 or price < 0:
            raise ValueError(f'Quantity {quantity} must be positive non-zero and price {price} must be positive!')
        if 1 < min_quantity_pct < 0 or 1 < min_notional_pct < 0:
            raise ValueError(f'min_quantity_pct {min_quantity_pct} and min_notional_pct {min_notional_pct} both '
                             f'must be between 0..1 inclusive')

        # Set rounding function
        rfn = ceil if rounding == 'ceil' else floor if rounding == 'floor' else round

        # Applying LOT_SIZE.step_size filter if exist
        if self.filter_step_size:
            quantity = rfn(Decimal(quantity) / self.filter_step_size) * self.filter_step_size

        # Applying MIN_NOTIONAL.min_notional filter if exist and min_notional_pct rule if price is given
        if price and self.filter_min_notional and self.filter_step_size:
            price = Decimal(price)
            quote_quantity = quantity * price
            if self.filter_min_notional > quote_quantity >= Decimal(min_notional_pct) * self.filter_min_notional:
                quantity = ceil(self.filter_min_notional / price / self.filter_step_size) * self.filter_step_size
            elif quote_quantity < self.filter_min_notional:
                quantity = 0

        # Applying LOT_SIZE.min_quantity and LOT_SIZE.max_quantity filters if exist and min_quantity_pct rule
        if self.filter_min_quantity and quantity < self.filter_min_quantity:
            if quantity >= Decimal(min_quantity_pct) * self.filter_min_quantity:
                quantity = self.filter_min_quantity
            else:
                quantity = 0
        elif self.filter_max_quantity and quantity > self.filter_max_quantity:
            quantity = self.filter_max_quantity

        return return_as(quantity)



import unittest
from typing import Any, Dict

from amarket.base import SignalGeneratorBase
from pandas import DataFrame, Series


class SomeSignalGeneratorForTest(SignalGeneratorBase):
    """Test implementation of SignalGeneratorBase for testing purposes."""
    def __init__(self, param1: str = "default1", param2: int = 42):
        self.param1 = param1
        self.param2 = param2

    def featurize(self, dataframe: DataFrame) -> DataFrame:
        return DataFrame([], columns=['test_col1', 'test_col2', 'mixed_col3'])

    def generate_long_signals(self, dataframe: DataFrame) -> Series: pass
    def generate_short_signals(self, dataframe: DataFrame) -> Series: pass

    @classmethod
    def get_column_to_parameters_mapping(cls) -> dict[str, list[str]]:
        return {'test_col1': ['param1'], 'test_col2': ['param2'], 'mixed_col3': ['param1', 'param2']}

    @classmethod
    def get_parameter_to_columns_mapping(cls) -> dict[str, list[str]]:
        return {'param1': ['test_col1', 'mixed_col3'], 'param2': ['test_col2', 'mixed_col3']}


class TestSignalGeneratorBase(unittest.TestCase):
    """Test cases for SignalGeneratorBase class."""

    def setUp(self):
        self.dataframe = DataFrame()
        self.signal_generator = SomeSignalGeneratorForTest()
        self.expected_default_config = {'param1': 'default1', 'param2': 42}

    def test_config_methods(self):
        # Test that get_default_config returns correct default values
        default_config = SomeSignalGeneratorForTest.get_default_parameters()
        actual_config = self.signal_generator.get_actual_parameters()
        self.assertDictEqual(default_config, self.expected_default_config)
        self.assertDictEqual(default_config, actual_config)

        # Test that validate_parameters returns True for valid configuration
        self.assertTrue(SomeSignalGeneratorForTest.validate_parameters({'param1': 'test_value', 'param2': 100}))

        # Test that validate_parameters returns False for invalid configuration
        invalid_config: Dict[str, Any] = {
            'param1': 'test_value'
            # Missing param2
        }
        self.assertFalse(SomeSignalGeneratorForTest.validate_parameters(invalid_config))

        # Test that validate_parameters returns False for empty configuration
        self.assertFalse(SomeSignalGeneratorForTest.validate_parameters({}))

    def test_hash_and_equality_magic_methods(self):
        # Test that hash method works correctly
        obj2 = SomeSignalGeneratorForTest()
        obj3 = SomeSignalGeneratorForTest(param1="different")

        # Objects with same parameters should have same hash
        self.assertEqual(hash(self.signal_generator), hash(obj2))
        self.assertIsNot(self.signal_generator, obj2)  # but they are not the same object

        # Objects with different parameters should have different hash
        self.assertNotEqual(hash(self.signal_generator), hash(obj3))
        self.assertIsNot(self.signal_generator, obj3)  # but they are not the same object

        # Test equality
        self.assertEqual(self.signal_generator, obj2)
        self.assertNotEqual(self.signal_generator, obj3)

        # Test that objects are equal to themselves
        self.assertEqual(self.signal_generator, self.signal_generator)

        # Test that different types are not equal
        self.assertNotEqual(self.signal_generator, "not_a_signal_generator")

    def test_repr_method(self):
        # Test that repr method works correctly
        repr_str = repr(self.signal_generator)
        self.assertIn("SomeSignalGeneratorForTest", repr_str)
        self.assertIn("param1='default1'", repr_str)
        self.assertIn("param2=42", repr_str)

        # Test with different parameters
        obj2 = SomeSignalGeneratorForTest(param1="custom", param2=100)
        repr_str2 = repr(obj2)
        self.assertIn("param1='custom'", repr_str2)
        self.assertIn("param2=100", repr_str2)

    def test_feature_mapping_methods(self):
        # Test get_column_to_parameter_mapping
        feature_columns = self.signal_generator.featurize(self.dataframe).columns.to_list()
        col_to_param = self.signal_generator.get_column_to_parameters_mapping()
        self.assertEqual(col_to_param,
                         {'test_col1': ['param1'], 'test_col2': ['param2'], 'mixed_col3': ['param1', 'param2']})
        self.assertListEqual(feature_columns, list(col_to_param.keys()))

        # Test get_parameter_to_column_mapping
        param_to_col = self.signal_generator.get_parameter_to_columns_mapping()
        self.assertEqual(param_to_col, {'param1': ['test_col1', 'mixed_col3'], 'param2': ['test_col2', 'mixed_col3']})

        # Test get_feature_info
        info = self.signal_generator.get_feature_info()
        self.assertIn('columns', info)
        self.assertIn('parameters', info)
        self.assertIn('column_to_parameters', info)
        self.assertIn('parameter_to_columns', info)

    def test_validate_feature_columns(self):
        valid_columns = ['test_col1', 'test_col2', 'mixed_col3']  # test with valid columns
        self.assertTrue(SomeSignalGeneratorForTest.validate_feature_columns(valid_columns))

        invalid_columns = ['test_col1', 'test_col2']  # missing 'mixed_col3'
        self.assertFalse(SomeSignalGeneratorForTest.validate_feature_columns(invalid_columns))

        extra_columns = ['test_col1', 'test_col2', 'mixed_col3', 'extra_col']  # extra columns (should still be valid)
        self.assertTrue(SomeSignalGeneratorForTest.validate_feature_columns(extra_columns))

        self.assertFalse(SomeSignalGeneratorForTest.validate_feature_columns([]))  # test with empty columns

        # Test with invalid input type - non-iterable (should not crash)
        # This would cause a TypeError when trying to convert to set
        with self.assertRaises(TypeError):
            SomeSignalGeneratorForTest.validate_feature_columns(42)

    def test_validate(self):
        result = self.signal_generator.validate()
        self.assertIsInstance(result, dict)
        self.assertIn("success", result)
        self.assertIn("errors", result)
        self.assertTrue(result['success'])
        self.assertTrue(result['success'])
        self.assertEqual(result["errors"], [])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Unit tests for Featurizer class"""
import unittest

import numpy as np
import pandas as pd
from amarket.ohlc import Featurizer
from amarket.test_data import DATAFRAMES


class TestFeaturizer(unittest.TestCase):
    def setUp(self):
        self.data = DATAFRAMES['BTCUSDT_1d']
        self.all_features = {'net_return', 'log_return', 'gross_return', 'gross_return_from_log_return', 'atr', 'natr',
                             'prev', 'atrp', 'rsi', 'sma', 'diff', 'div', 'mult', 'marubozu', 'diff_extrema', 'macd',
                             'wf', 'future', 'ma_cross'}
        self.featurizer = Featurizer(pd.DataFrame(self.data))

    def test_initialization(self):
        self.assertTrue(hasattr(self.featurizer, 'original'))
        self.assertTrue(hasattr(self.featurizer, 'applied'))
        self.assertTrue(hasattr(self.featurizer, 'known_features'))
        self.assertTrue(hasattr(self.featurizer, 'cols'))
        self.assertTrue(hasattr(self.featurizer, 'used'))

        self.assertIsInstance(self.featurizer.original, pd.DataFrame)
        self.assertIsInstance(self.featurizer.applied, pd.DataFrame)
        self.assertIsInstance(self.featurizer.known_features, set)
        self.assertIsInstance(self.featurizer.cols, list)
        self.assertIsInstance(self.featurizer.used, list)

        self.assertGreater(len(self.featurizer.known_features), 0)
        self.assertEqual(len(self.featurizer), len(self.data))
        self.assertListEqual(self.featurizer.cols, [])
        self.assertListEqual(self.featurizer.used, [])

        self.assertTrue(self.featurizer.original.equals(self.featurizer.applied))
        self.assertSetEqual(self.all_features, self.featurizer.get_known_features())
        self.assertSetEqual(self.all_features, self.featurizer.known_features)

    def test_apply_features(self):
        features = (self.featurizer.known_features -
                    {'div', 'mult', 'diff_extrema', 'diff', 'prev', 'future', 'net_return', 'log_return', 'ma_cross'})
        cols, used = self.featurizer.apply_features(['net_return', {'feature': 'log_return', 'use': None}]
                                                    + list(features))
        self.assertIsInstance(cols, list)
        self.assertIsInstance(used, list)
        self.assertGreater(len(cols), len(used))
        self.assertIn('net_return', cols)
        self.assertIn('net_return', used)
        self.assertIn('log_return', cols)
        self.assertNotIn('log_return', used)

    def test_validate(self):
        data = self.data.copy()
        self.featurizer.validate(pd.DataFrame(self.data))  # no exception
        with self.assertRaises(ValueError):  # exception
            del data['close']
            self.featurizer.validate(pd.DataFrame(data))
        with self.assertRaises(ValueError):  # exception
            self.featurizer.validate(pd.DataFrame({"date": [1, 2], "open": [10, 20]}))

    def test_net_return_feature(self):
        feats = ['net_return',
                 {'feature': 'net_return', 'kwargs': {'period': 2}},
                 {'feature': 'net_return', 'kwargs': {'period': 5}},]

        cols, used = self.featurizer.apply_features(feats)
        self.assertListEqual(cols, used)

        period = 1
        column = 'close'
        for f in feats:
            if isinstance(f, dict) and 'kwargs' in f and 'period' in f['kwargs']:
                period = f['kwargs']['period']
                f = f['feature']
            final = self.featurizer.applied.loc[period, column]
            initial = self.featurizer.applied.loc[period - 1, column]
            result = self.featurizer.applied.loc[period, f]
            self.assertAlmostEqual(result, (final-initial)/initial)

    def test_gross_return_features(self):
        self.featurizer.apply_features(['net_return', 'log_return', 'gross_return', 'gross_return_from_log_return'])
        self.assertTrue(np.isclose(self.featurizer.applied['gross_return'],
                                   self.featurizer.applied.gross_return_from_log_return, equal_nan=True).all())
        self.assertIsNone(pd.testing.assert_series_equal(self.featurizer.applied['gross_return'],
                                                         self.featurizer.applied.gross_return_from_log_return,
                                                         check_names=False))

    def test_rsi_feature(self):
        # Reference RSI values from TradingView, obtained manually
        expected_rsi = pd.DataFrame({
            'date': [f"2023-07-0{_}" for _ in range(1, 8)],  # 01-07 Jul, 2023
            'rsi': [66.81, 67, 70.16, 65.29, 62.18, 55.53, 59.01],
        })
        expected_rsi['date'] = pd.to_datetime(expected_rsi['date'], utc=True)
        cols, used = self.featurizer.apply_features(['rsi'])

        self.assertIsNone(pd.testing.assert_frame_equal(
            expected_rsi, self.featurizer.applied[['date'] + cols].iloc[-7:].reset_index(drop=True), atol=0.01))

        cols, used = Featurizer(self.data.iloc[-140:]).apply_features(['rsi'])
        self.assertIsNone(pd.testing.assert_frame_equal(
            expected_rsi, self.featurizer.applied[['date'] + cols].iloc[-7:].reset_index(drop=True), atol=0.01))

    def test_prev_feature(self):
        features = ['rsi', 'prev_rsi_1', 'prev_rsi_7', 'prev_diff_pct(close-sma_14)_1', 'prev_diff_pct(close-sma_14)_7',
                    {'feature': 'sma', 'kwargs': {'period': 14}},
                    {'feature': 'diff', 'args': ['close', 'sma_14']}]
        self.featurizer.apply_features(features)
        self.assertTrue(np.isclose(self.featurizer.applied['rsi'].shift(1),
                                   self.featurizer.applied['prev_rsi_1'], equal_nan=True).all())
        self.assertTrue(np.isclose(self.featurizer.applied['rsi'].shift(7), self.featurizer.applied['prev_rsi_7'],
                                   equal_nan=True).all())
        self.assertTrue(np.isclose(self.featurizer.applied['diff_pct(close-sma_14)'].shift(1),
                                   self.featurizer.applied['prev_diff_pct(close-sma_14)_1'], equal_nan=True).all())
        self.assertTrue(np.isclose(self.featurizer.applied['diff_pct(close-sma_14)'].shift(7),
                                   self.featurizer.applied['prev_diff_pct(close-sma_14)_7'], equal_nan=True).all())

    def test_future_feature(self):
        features = ["log_return", "future_log_return_1"]
        self.featurizer.apply_features(features)
        self.assertTrue(np.isclose(self.featurizer.applied['log_return'].shift(-1),
                                   self.featurizer.applied['future_log_return_1'], equal_nan=True).all())

    def test_sma_feature(self):
        # Reference SMA values from TradingView, obtained manually
        for period, data in {7: [16791.6, 16864.62, 16936.81, 17046.17, 17202.3, 17490.12, 17915.74],
                             14: [17640.2, 17944.11, 18266.45, 18585, 18858.36, 19161.2, 19569.53, 19986.67]}.items():
            cols, used = self.featurizer.apply_features([{'feature': 'sma', 'kwargs': {'period': period}}])
            self.assertListEqual(cols, used)
            name = cols[0]
            start = self.data.iloc[period-1, self.data.columns.get_loc('date')]
            expected_sma = pd.Series(data, name=name, index=pd.date_range(start=start, periods=len(data), freq='1D'))
            sma = self.featurizer.applied.set_index('date', drop=True)[name].loc[start:expected_sma.index.max()]
            self.assertTrue(np.isclose(expected_sma, sma).all())
            self.assertIsNone(pd.testing.assert_series_equal(
                expected_sma, sma, check_freq=False, check_names=False, rtol=0.01))
            self.assertTupleEqual((name,), self.featurizer.feature_sma(period=period))

    def test_diff_feature(self):
        expected_cols = ('diff(close-sma_14)', 'diff_pct(close-sma_14)')
        cols, used = self.featurizer.apply_features([{'feature': 'sma', 'kwargs': {'period': 14}},
                                                     {'feature': 'diff', 'args': ['close', 'sma_14']}])
        self.assertListEqual(['sma_14'] + list(expected_cols), cols)
        self.assertTupleEqual(expected_cols, self.featurizer.feature_diff('close', 'sma_14'))

    def test_div_feature(self):
        expected_cols = ('div(high/close)',)
        cols, used = self.featurizer.apply_features([{'feature': 'div', 'args': ['high', 'close']}])
        self.assertListEqual(list(expected_cols), cols)
        self.assertTupleEqual(expected_cols, self.featurizer.feature_div('high', 'close'))

    def test_mult_feature(self):
        expected_cols = ('mult(high*close)',)
        cols, used = self.featurizer.apply_features([{'feature': 'mult', 'args': ['high', 'close']}])
        self.assertListEqual(list(expected_cols), cols)
        self.assertTupleEqual(expected_cols, self.featurizer.feature_mult('high', 'close'))

    def test_marubozu_feature(self):
        expected_cols = ('marubozu',)
        cols, used = self.featurizer.apply_features(['marubozu'])
        self.assertListEqual(list(expected_cols), cols)
        self.assertTupleEqual(expected_cols, self.featurizer.feature_marubozu())

    def test_wf_feature(self):
        expected_cols = ('wf_bear', 'wf_bull', 'wf_bear_rt', 'wf_bull_rt')
        cols, used = self.featurizer.apply_features(['wf'])
        self.assertListEqual(list(expected_cols), cols)
        self.assertTupleEqual(expected_cols, self.featurizer.feature_wf())
        self.assertTrue(self.featurizer.applied['wf_bear'].iloc[-2:].eq(-1).all())
        self.assertTrue(self.featurizer.applied['wf_bull'].iloc[-2:].eq(-1).all())
        self.assertTrue(self.featurizer.applied['wf_bear_rt'].ne(-1).all())
        self.assertTrue(self.featurizer.applied['wf_bull_rt'].ne(-1).all())

        use = 'wf_bear_rt'
        cols, used = self.featurizer.apply_features([{'feature': 'wf', 'use': use}])
        self.assertListEqual(list(expected_cols), cols)
        self.assertListEqual([use], used)

        use = ['wf_bull_rt', 'wf_bear_rt']
        cols, used = self.featurizer.apply_features([{'feature': 'wf', 'use': use}])
        self.assertListEqual(list(expected_cols), cols)
        self.assertListEqual(use, used)

        cols, used = self.featurizer.apply_features([{'feature': 'wf', 'use': None}])
        self.assertListEqual(list(expected_cols), cols)
        self.assertListEqual([], used)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Contains SymbolAccountMixin and SymbolAccountError classes that implements basic symbol representation with balances
for base and quote currencies"""


class SymbolAccountError(Exception):
    """SymbolAccountMixin error"""


class SymbolAccountMixin:
    @property
    def base_quantity(self): return self._base_quantity

    @property
    def quote_quantity(self): return self._quote_quantity

    def __init__(self, *args, base_quantity=0, quote_quantity=0, **kwargs):
        if base_quantity < 0 or quote_quantity < 0:
            raise SymbolAccountError('Base and quote currencies quantity must be positive or zero!')
        self._base_quantity = base_quantity
        self._quote_quantity = quote_quantity
        if self.__class__.mro() != [self.__class__, object]:  # call parent(s).__init__ if it not 'object' only
            super().__init__(*args, **kwargs)



import unittest
from decimal import Decimal
from time import time

from amarket.constants import (
    ORDER_TYPE_LIMIT,
    ORDER_TYPE_MARKET,
    SIDE_BUY,
    SIDE_SELL,
    MarketType,
)
from amarket.interfaces.i_order import IOrder
from amarket.local.local_market import LocalMarket
from amarket.local.local_order import LocalOrder
from test_utils import check_primitive_types


class TestLocalOrder(unittest.TestCase):
    """Test cases for the LocalOrder class."""

    def setUp(self):
        self.market = LocalMarket(exchange='Exchange_1', market_type="spot", symbol="BTCUSDT",
                                  base_ticker="BTC", quote_ticker="USDT", lot_precision=4, price_precision=2)
        self.order = LocalOrder(market=self.market, order_id=1001, owner='owner#1',
                                client_order_id="test_client_1", side=SIDE_BUY, order_type="LIMIT", status="NEW",
                                quantity=1.5, price=100.0, create_time=int(time()), update_time=int(time()))
        self.validation = self.order.validate()

        # Create valid order dictionaries
        self.valid_limit_order = {
            'market': self.market,
            'order_id': 123456,
            'owner': 'owner#1',
            'client_order_id': 'some_client_order_id_123456',
            'side': SIDE_BUY,
            'order_type': ORDER_TYPE_LIMIT,
            'status': 'NEW',
            'quantity': Decimal('1.0'),
            'price': Decimal('30000.0'),
            'create_time': int(time()),
            'update_time': int(time()),
            'provider_data': {'some_key': 'some_value'}
        }
        self.valid_market_order = {
            'market': self.market,
            'order_id': 654321,
            'owner': 'owner#1',
            'client_order_id': 'some_client_order_id_654321',
            'side': SIDE_SELL,
            'order_type': ORDER_TYPE_MARKET,
            'status': 'FILLED',
            'quantity': Decimal('0.5'),
            'price': None,  # Price can be None for MARKET orders
            'create_time': int(time()),
            'update_time': int(time()),
        }

    def test_initialization(self):
        """Test successful creation of a LocalOrder instance"""
        self.assertIsInstance(self.order, LocalOrder)
        self.assertIsInstance(self.order, IOrder)
        self.assertIn('success', self.validation)
        self.assertIn('errors', self.validation)
        self.assertTrue(self.validation['success'])
        self.assertEqual(self.validation['errors'], [])
        self.assertTrue(self.order.is_buy)

        # Test properties for exchange, market_type, and symbol
        self.assertEqual(self.order.exchange, self.market.exchange)
        self.assertEqual(self.order.market_type, MarketType.get_market_type_code(self.market.market_type))
        self.assertEqual(self.order.symbol, self.market.symbol)

    def test_to_dict(self):
        """Test the to_dict() method"""
        order_dict = self.order.to_dict()

        # Verify all expected fields are present and values match the order's properties
        expected_fields = ["market", "order_id", "owner", "client_order_id", "side",
                           "order_type", "status", "quantity", "price", "create_time", "update_time",
                           "check_time", "provider_data", "history", "exchange", "market_type", "symbol"]
        for field in expected_fields:
            self.assertIn(field, order_dict)
            if field == 'market':
                self.assertDictEqual(order_dict['market'], self.order.market.to_dict())
            else:
                self.assertEqual(order_dict[field], getattr(self.order, field))

        # Test that to_dict() returns only primitive types and valid nested structures
        result = self.order.to_dict()
        check_primitive_types(result)

    def test_validate(self):
        # Test validation over some missing or invalid properties
        invalid_props = {'market': None, 'order_id': 0, 'status': None}
        for prop, value in invalid_props.items():
            init_value = getattr(self.order, prop)
            setattr(self.order, f'_{prop}', value)
            result = self.order.validate()
            self.assertFalse(result["success"])
            self.assertIn(f"Invalid {prop}", result["errors"][0])
            self.assertEqual(len(result['errors']), 1)
            setattr(self.order, f'_{prop}', init_value)  # restore

        # Test valid orders creation
        for test_order in (self.valid_limit_order, self.valid_market_order):
            order = LocalOrder(**test_order)
            validation_result = order.validate()
            self.assertTrue(validation_result['success'],
                            f"Valid LIMIT order should pass validation, but validation result is: {validation_result}")
            self.assertEqual(len(validation_result['errors']), 0, "No errors expected for valid LIMIT order")

        # Test missing fields
        for field in ['market', 'order_id', 'client_order_id', 'side', 'order_type', 'status', 'quantity',
                      'create_time', 'update_time']:
            for test_order in (self.valid_limit_order, self.valid_market_order):
                order = LocalOrder(**test_order)
                setattr(order,  f'_{field}', None)
                validation_result = order.validate()
                self.assertFalse(validation_result['success'], f"Order missing {field} should fail validation")
                self.assertIn(f"Invalid {field}", validation_result['errors'][0])

        # Test invalid symbol
        order = LocalOrder(**self.valid_limit_order)
        market_symbol = order.market.symbol
        order.market._symbol = ''
        validation_result = order.validate()
        self.assertFalse(validation_result['success'], "Empty symbol should fail validation")
        self.assertIn("Invalid symbol: must be a non-empty alphanumeric string without spaces",
                      validation_result['errors'])

        order.market._symbol = 12345
        validation_result = order.validate()
        self.assertFalse(validation_result['success'], "Non-string symbol should fail validation")
        self.assertIn("Invalid symbol: must be a non-empty alphanumeric string without spaces",
                      validation_result['errors'])
        order.market._symbol = market_symbol  # restore the market symbol

        # Test invalid side
        order = LocalOrder(**self.valid_limit_order)
        order._side = 0  # Not SIDE_BUY or SIDE_SELL
        validation_result = order.validate()
        self.assertFalse(validation_result['success'], "Invalid side should fail validation")
        self.assertIn(f"Invalid side: must be {SIDE_BUY} or {SIDE_SELL}", validation_result['errors'])

        # Test invalid order type
        order = LocalOrder(**self.valid_limit_order)
        order._order_type = 'INVALID'
        validation_result = order.validate()
        self.assertFalse(validation_result['success'], "Invalid order type should fail validation")
        self.assertIn(f"Invalid order_type: must be '{ORDER_TYPE_LIMIT}' or '{ORDER_TYPE_MARKET}'",
                      validation_result['errors'])

        # Test invalid order ID
        order = LocalOrder(**self.valid_limit_order)
        for order_id, msg in ((0, 'Zero order ID should fail validationn'),
                              (-1, 'Negative order ID should fail validation'),
                              ('12345', 'Non-integer order ID should fail validation')):
            order._order_id = order_id
            validation_result = order.validate()
            self.assertFalse(validation_result['success'], msg)
            self.assertIn("Invalid order_id: must be a positive integer", validation_result['errors'])

        # Test invalid quantity
        order = LocalOrder(**self.valid_limit_order)
        for quantity, msg in ((Decimal('0'), 'Zero quantity should fail validation'),
                              (Decimal('-1.0'), 'Negative quantity should fail validation'),
                              ('1.0', 'Non-Decimal quantity should fail validation')):
            order._quantity = quantity
            validation_result = order.validate()
            self.assertFalse(validation_result['success'], msg)
            self.assertIn("Invalid quantity: must be a positive Decimal", validation_result['errors'])

        # Test invalid price for LIMIT order
        order = LocalOrder(**self.valid_limit_order)
        for price, msg in ((Decimal('0'), 'Zero price should fail validation'),
                           (Decimal('-1.0'), 'Negative price should fail validation'),
                           ('30000', 'Non-Decimal price should fail validation')):
            order._price = price
            validation_result = order.validate()
            self.assertFalse(validation_result['success'], msg)
            self.assertIn("Invalid price for LIMIT order: must be a positive Decimal", validation_result['errors'])

        # Test invalid status
        order._price = Decimal('30000')  # Reset to valid value
        order._status = ''
        validation_result = order.validate()
        self.assertFalse(validation_result['success'], "Empty status should fail validation")
        self.assertIn("Invalid status: must be a non-empty string", validation_result['errors'])

        order._status = 12345
        validation_result = order.validate()
        self.assertFalse(validation_result['success'], "Non-string status should fail validation")
        self.assertIn("Invalid status: must be a non-empty string", validation_result['errors'])

        # Test invalid provider_data
        order._provider_data = 'not a dict'
        validation_result = order.validate()
        self.assertFalse(validation_result['success'], "Non-dict provider_data should fail validation")
        self.assertIn("Invalid provider_data: must be a dictionary or None", validation_result['errors'])

    def test_is_buy_and_is_sell_properties(self):
        """Comprehensive test for is_buy and is_sell properties with various scenarios"""
        # Test case 1: SIDE_BUY
        buy_order = LocalOrder(market=self.market, order_id=1, owner='owner#1',
                               client_order_id="test_client_1", side=SIDE_BUY, order_type="LIMIT",
                               status="NEW", quantity=1.5, price=100.0,
                               create_time=int(time()), update_time=int(time()))
        self.assertTrue(buy_order.is_buy, "is_buy should be True when side is SIDE_BUY")
        self.assertFalse(buy_order.is_sell, "is_sell should be False when side is SIDE_BUY")

        # Test case 2: SIDE_SELL
        sell_order = LocalOrder(market=self.market, order_id=2, owner='owner#2',
                                client_order_id="test_client_2", side=SIDE_SELL, order_type="LIMIT",
                                status="NEW", quantity=1.5, price=100.0,
                                create_time=int(time()), update_time=int(time()))
        self.assertTrue(sell_order.is_sell, "is_sell should be True when side is SIDE_SELL")
        self.assertFalse(sell_order.is_buy, "is_buy should be False when side is SIDE_SELL")

        # Test case 3: Invalid side
        invalid_order = LocalOrder(market=self.market, order_id=3, owner='owner#3',
                                   client_order_id="test_client_3", side=0,  # Invalid side
                                   order_type="LIMIT", status="NEW", quantity=1.5,
                                   price=100.0, create_time=int(time()), update_time=int(time()))
        # Both properties should be False for invalid side
        self.assertFalse(invalid_order.is_buy, "is_buy should be False for invalid side")
        self.assertFalse(invalid_order.is_sell, "is_sell should be False for invalid side")

        # Validation should fail
        validation_result = invalid_order.validate()
        self.assertFalse(validation_result['success'], "Validation should fail for invalid side")
        self.assertIn("Invalid side: must be 1 or -1", validation_result['errors'])

        # Test case 4: Edge cases
        for side_value in [SIDE_BUY, SIDE_SELL, 0, 2, -2]:
            edge_order = LocalOrder(market=self.market, order_id=4, owner='owner#4',
                                    client_order_id="test_client_4", side=side_value,
                                    order_type="LIMIT", status="NEW", quantity=1.5,
                                    price=100.0, create_time=int(time()), update_time=int(time()))

            # Check property values based on side
            if side_value == SIDE_BUY:
                self.assertTrue(edge_order.is_buy, f"is_buy should be True for side={side_value}")
                self.assertFalse(edge_order.is_sell, f"is_sell should be False for side={side_value}")
            elif side_value == SIDE_SELL:
                self.assertTrue(edge_order.is_sell, f"is_sell should be True for side={side_value}")
                self.assertFalse(edge_order.is_buy, f"is_buy should be False for side={side_value}")
            else:
                self.assertFalse(edge_order.is_buy, f"is_buy should be False for invalid side={side_value}")
                self.assertFalse(edge_order.is_sell, f"is_sell should be False for invalid side={side_value}")

                # Validation should fail for invalid sides
                validation_result = edge_order.validate()
                self.assertFalse(validation_result['success'], f"Validation should fail for side={side_value}")
                self.assertIn("Invalid side: must be 1 or -1", validation_result['errors'])


if __name__ == "__main__":
    unittest.main()



from .asset_set import AssetSet, AssetSetError
from .position import Position, PositionError, PositionLimit
from .symbol_account_mixin import SymbolAccountError, SymbolAccountMixin
from .symbol_mixin import SymbolError, SymbolMixin

__all__ = ['PositionError', 'Position', 'PositionLimit', 'SymbolError', 'SymbolMixin',
           'SymbolAccountError', 'SymbolAccountMixin', 'AssetSetError', 'AssetSet']



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Tests some logic in constants module"""
import unittest

from amarket.constants import SIDE_BUY, SIDE_SELL, MarketType, Side


class TestConstants(unittest.TestCase):
    def test_market_type(self):
        # Test is_spot methods
        self.assertTrue(MarketType.is_spot(10))
        self.assertFalse(MarketType.is_spot(20))
        self.assertTrue(MarketType.is_spot('spot'))
        self.assertFalse(MarketType.is_spot('futures'))

        # Test is_futures methods
        self.assertTrue(MarketType.is_futures(20))
        self.assertFalse(MarketType.is_futures(10))
        self.assertTrue(MarketType.is_futures('futures'))
        self.assertFalse(MarketType.is_futures('spot'))

        # Test is_cross_margin methods
        self.assertTrue(MarketType.is_cross_margin(30))
        self.assertFalse(MarketType.is_cross_margin(40))
        self.assertTrue(MarketType.is_cross_margin('cross'))
        self.assertFalse(MarketType.is_cross_margin('isolated'))

        # Test is_isolated_margin methods
        self.assertTrue(MarketType.is_isolated_margin(40))
        self.assertFalse(MarketType.is_isolated_margin(30))
        self.assertTrue(MarketType.is_isolated_margin('isolated'))
        self.assertFalse(MarketType.is_isolated_margin('cross'))

        # Test possible_types
        result = MarketType.possible_types()
        for s in ('spot', 'futures', 'cross', 'isolated'):
            self.assertIn(s, result)

        # Test possible_codes
        result = MarketType.possible_codes()
        for v in (10, 20, 30, 40):
            self.assertIn(v, result)

        # Test get_market_type_code methods
        self.assertEqual(MarketType.get_market_type_code(10), 10)
        self.assertEqual(MarketType.get_market_type_code('spot'), 10)
        self.assertEqual(MarketType.get_market_type_code(20), 20)
        self.assertEqual(MarketType.get_market_type_code('futures'), 20)
        self.assertEqual(MarketType.get_market_type_code(30), 30)
        self.assertEqual(MarketType.get_market_type_code('cross'), 30)
        self.assertEqual(MarketType.get_market_type_code(40), 40)
        self.assertEqual(MarketType.get_market_type_code('isolated'), 40)

        # Test get_market_type methods
        self.assertEqual(MarketType.get_market_type(10), 'spot')
        self.assertEqual(MarketType.get_market_type('spot'), 'spot')
        self.assertEqual(MarketType.get_market_type(20), 'futures')
        self.assertEqual(MarketType.get_market_type('futures'), 'futures')
        self.assertEqual(MarketType.get_market_type(30), 'cross')
        self.assertEqual(MarketType.get_market_type('cross'), 'cross')
        self.assertEqual(MarketType.get_market_type(40), 'isolated')
        self.assertEqual(MarketType.get_market_type('isolated'), 'isolated')

        # Test invalid market type
        with self.assertRaises(ValueError):
            MarketType.get_market_type(999)

    def test_side_type(self):
        # Test is_buy method
        self.assertTrue(Side.is_buy(1))
        self.assertTrue(Side.is_buy('1'))
        self.assertFalse(Side.is_buy(-1))
        self.assertTrue(Side.is_buy('BUY'))
        self.assertTrue(Side.is_buy('LONG'))
        self.assertFalse(Side.is_buy('SELL'))
        self.assertFalse(Side.is_buy('SHORT'))

        # Test is_sell method
        self.assertTrue(Side.is_sell(-1))
        self.assertTrue(Side.is_sell('-1'))
        self.assertFalse(Side.is_sell(1))
        self.assertTrue(Side.is_sell('SELL'))
        self.assertTrue(Side.is_sell('SHORT'))
        self.assertFalse(Side.is_sell('BUY'))
        self.assertFalse(Side.is_sell('LONG'))

        # Test possible_types
        result = Side.possible_types()
        for s in ('BUY', 'SELL'):
            self.assertIn(s, result)

        # Test possible_codes
        result = Side.possible_codes()
        for v in (1, -1):
            self.assertIn(v, result)

        # Test get_code methods
        self.assertEqual(Side.get_code(1), 1)
        self.assertEqual(Side.get_code('1'), 1)
        self.assertEqual(Side.get_code('BUY'), 1)
        self.assertEqual(Side.get_code(Side.BUY), 1)
        self.assertEqual(Side.get_code('LONG'), 1)
        self.assertEqual(Side.get_code(Side.LONG), 1)
        self.assertEqual(Side.get_code(-1), -1)
        self.assertEqual(Side.get_code('-1'), -1)
        self.assertEqual(Side.get_code('SELL'), -1)
        self.assertEqual(Side.get_code(Side.SELL), -1)
        self.assertEqual(Side.get_code('SHORT'), -1)
        self.assertEqual(Side.get_code(Side.SHORT), -1)

        self.assertEqual(Side.BUY_CODE, Side.LONG)
        self.assertEqual(Side.SELL_CODE, Side.SHORT)

        # Test invalid side type
        with self.assertRaises(ValueError):
            Side.get_code(999)

        # Test constants are the same
        self.assertEqual(Side.BUY_CODE, SIDE_BUY)
        self.assertEqual(Side.SELL_CODE, SIDE_SELL)


if __name__ == '__main__':
    unittest.main()



import unittest

import numpy as np
from amarket.portfolio import Portfolio, PortfolioError


class TestPortfolio(unittest.TestCase):
    def setUp(self):
        self.wealth = 10000
        self.prices = [37175, 1.0]
        self.weights = [0.7, 0.3]
        self.tickers = ['BTC', 'USDT']
        self.currency = 'USDT'
        self.n = len(self.prices)
        self.portfolio = [self.wealth*self.weights[0]/self.prices[0], self.wealth*self.weights[1]/self.prices[1]]
        self.value = self.portfolio[0] * self.prices[0] + self.portfolio[1] * self.prices[1]
        self.p = Portfolio(self.wealth, self.prices, weights=self.weights, tickers=self.tickers, currency=self.currency)
        self.p_1n = Portfolio(self.wealth, self.prices, tickers=self.tickers, currency=self.currency)
        self.p_usdt = Portfolio(self.wealth, [0, 1], weights=[0, 1], tickers=self.tickers, currency=self.currency)

    def test_initialization(self):
        # Test properties of self.p
        self.assertEqual(self.p.initial_wealth, 10000)
        self.assertEqual(self.p.wealth, 10000)
        self.assertEqual(self.p.currency, self.currency)
        self.assertEqual(self.p.n, self.n)
        self.assertSequenceEqual(self.p.tickers, self.tickers)
        self.assertEqual(len(self.p.history), 1)
        self.assertSequenceEqual(self.p.prices.tolist(), self.prices)  # prices 1 method
        self.assertTrue(np.allclose(self.p.prices, self.prices))  # prices 2 method
        self.assertSequenceEqual(self.p.weights.tolist(), self.weights)  # weights 1 method
        self.assertTrue(np.allclose(self.p.weights, self.weights))  # weights 2 method
        self.assertEqual(sum(self.p.weights), 1)
        self.assertSequenceEqual(self.p.portfolio.tolist(), self.portfolio)  # portfolio 1 method
        self.assertTrue(np.allclose(self.p.portfolio, self.portfolio))  # portfolio 2 method

        # Test properties of self.p_1n
        self.assertEqual(self.p_1n.initial_wealth, self.wealth)
        self.assertEqual(self.p_1n.wealth, self.wealth)
        self.assertEqual(self.p_1n.n, self.n)
        self.assertEqual(len(self.p_1n.history), 1)
        weights = np.empty(self.n)
        weights.fill(1/self.n)
        self.assertSequenceEqual(self.p_1n.weights.tolist(), weights.tolist())
        self.assertTrue(np.allclose(self.p_1n.weights, weights))
        self.assertEqual(sum(self.p_1n.weights), 1)
        portfolio = [self.wealth*weights[0]/self.prices[0], self.wealth*weights[1]/self.prices[1]]
        self.assertSequenceEqual(self.p_1n.portfolio.tolist(), portfolio)  # portfolio 1 method
        self.assertTrue(np.allclose(self.p_1n.portfolio, portfolio))  # portfolio 2 method

        # Test properties of self.p_usdt
        self.assertEqual(self.p_usdt.initial_wealth, self.wealth)
        self.assertEqual(self.p_usdt.wealth, self.wealth)
        self.assertEqual(self.p_usdt.n, self.n)
        self.assertEqual(len(self.p_usdt.history), 1)
        self.assertSequenceEqual(self.p_usdt.prices.tolist(), [1, 1])
        self.assertSequenceEqual(self.p_usdt.weights.tolist(), [0, 1])
        self.assertEqual(sum(self.p_usdt.weights), 1)
        self.assertSequenceEqual(self.p_usdt.portfolio.tolist(), [0, self.wealth])

        # Test raising exceptions for wrong portfolios
        for prices, weights in zip([[0, 0], [0, 0], [1, 40], [100, 22]], [[0, 0], [0, 1], [0.3, 0.9], [0.3, 0.2, 0.5]]):
            with self.assertRaises(PortfolioError):
                Portfolio(self.wealth, prices, weights=weights)
        with self.assertRaises(PortfolioError):
            Portfolio(0, self.prices, weights=self.weights)

    def test_value(self):
        self.assertEqual(self.p.value(), self.value)
        self.assertEqual(self.p.value(), self.wealth)
        self.assertEqual(self.p.value(), self.p.wealth)
        self.assertEqual(self.p.value(), self.p.initial_wealth)

        self.assertEqual(self.p_1n.value(), self.value)
        self.assertEqual(self.p_1n.value(), self.wealth)
        self.assertEqual(self.p_1n.value(), self.p_1n.wealth)
        self.assertEqual(self.p_1n.value(), self.p_1n.initial_wealth)

        self.assertEqual(self.p_usdt.value(), self.value)
        self.assertEqual(self.p_usdt.value(), self.wealth)
        self.assertEqual(self.p_usdt.value(), self.p_usdt.wealth)
        self.assertEqual(self.p_usdt.value(), self.p_usdt.initial_wealth)

    def test_rebalance(self):
        new_prices = [36500, 1]
        new_weights = [0.5, 0.5]
        self.p.rebalance(new_prices, weights=new_weights)
        self.assertEqual(len(self.p.history), 2)
        self.assertSequenceEqual(self.p.prices.tolist(), new_prices)  # prices 1 method
        self.assertTrue(np.allclose(self.p.prices, new_prices))  # prices 2 method
        self.assertSequenceEqual(self.p.weights.tolist(), new_weights)  # weights 1 method
        self.assertTrue(np.allclose(self.p.weights, new_weights))  # weights 2 method
        self.assertTrue(np.allclose(self.p.portfolio, [0.135245184, 4936.44923]))  # rebalanced portfolio

        self.p.rebalance([20000, 1], wealth=1000)  # rebalance depositing wealth
        self.assertEqual(len(self.p.history), 3)

        self.p.rebalance([25000, 1], wealth=-1000)  # rebalance withdrawing wealth
        self.assertEqual(len(self.p.history), 4)

        # Check that sells + buys = added_wealth on each history record
        for hr in self.p.history:
            diff = 0
            for i, amount in enumerate(hr['diff']):
                diff += amount * hr['prices'][i]
            self.assertAlmostEqual(hr['wealth'], diff)
            self.assertAlmostEqual(hr['wealth'], np.sum(hr['diff'] * hr['prices']))  # calculations with numpy

        with self.assertRaises(PortfolioError):
            self.p.rebalance([35000, 1], wealth=-10000000)  # try to withdraw more money than total wealth
        self.assertEqual(len(self.p.history), 4)

        self.p.rebalance([5000, 1])  # simulate that prices go very low
        self.assertLess(self.p.value(), self.p.wealth)  # and now value is less than invested wealth
        with self.assertRaises(PortfolioError):
            self.p.rebalance([10000, 1], wealth=-self.p.wealth)  # try to withdraw all invested wealth
        self.assertEqual(len(self.p.history), 5)


if __name__ == '__main__':
    unittest.main()
